# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Valerio Cosentino <valcos@bitergia.com>
#

import configparser
import json
import os
import random
import string
import unittest
import unittest.mock

import httpretty
import requests

from grimoire_elk.elastic import (ElasticSearch,
                                  ElasticError,
                                  logger)
from grimoire_elk.raw.git import GitOcean
from grimoire_elk.raw.kitsune import KitsuneOcean
from grimoire_elk.elastic_mapping import Mapping
from grimoirelab_toolkit.datetime import unixtime_to_datetime

CONFIG_FILE = 'tests.conf'
ES_DISTRIBUTION = 'elasticsearch'
OS_DISTRIBUTION = 'opensearch'


class MockElasticSearch(ElasticSearch):

    def __init__(self, url, index, major=None, distribution=None, mock_list_alias=False):
        self.requests = requests
        self.url = url
        self.index = index
        self.major = major
        self.distribution = distribution
        self.index_url = self.url + "/" + self.index
        self.mock_list_aliases = mock_list_alias

    def list_aliases(self):
        if not self.mock_list_aliases:
            return super().list_aliases()
        else:
            return []


def read_file(filename, mode='r'):
    with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), filename), mode) as f:
        content = f.read()
    return content


class TestElastic(unittest.TestCase):
    """Test Elastic"""

    target_index = "elastic_raw_test"

    @classmethod
    def setUpClass(cls):
        cls.config = configparser.ConfigParser()
        cls.config.read(CONFIG_FILE)
        cls.es_con = dict(cls.config.items('ElasticSearch'))['url']
        re = requests.get(cls.es_con, verify=False).json()
        cls.es_major = re['version']['number'].split('.')[0]
        cls.es_distribution = re['version'].get('distribution', ES_DISTRIBUTION)

    def tearDown(self):
        target_index_url = self.es_con + "/" + self.target_index
        requests.delete(target_index_url, verify=False)

    def test_init(self):
        """Test whether attributes are initialized"""

        elastic = ElasticSearch(self.es_con, self.target_index, GitOcean.mapping)
        self.assertEqual(elastic.url, self.es_con)
        self.assertEqual(elastic.index, self.target_index)
        self.assertEqual(elastic.index_url, self.es_con + "/" + self.target_index)
        self.assertIsNone(elastic.aliases)

    @httpretty.activate
    def test_check_instance_es_major_6(self):
        """Test whether the major version is correctly calculated for ElasticSearch 6.x"""

        body = """{
            "name" : "44BPNNH",
            "cluster_name" : "elasticsearch",
            "cluster_uuid" : "fIa1j8AQRfSrmuhTwb9a0Q",
            "version" : {
                "number" : "6.1.0",
                "build_hash" : "c0c1ba0",
                "build_date" : "2017-12-12T12:32:54.550Z",
                "build_snapshot" : false,
                "lucene_version" : "7.1.0",
                "minimum_wire_compatibility_version" : "5.6.0",
                "minimum_index_compatibility_version" : "5.0.0"
            },
            "tagline" : "You Know, for Search"
        }"""

        es_con = "http://es6.com"
        httpretty.register_uri(httpretty.GET,
                               es_con,
                               body=body,
                               status=200)

        major, distribution = ElasticSearch.check_instance(es_con, insecure=False)
        self.assertEqual(major, '6')
        self.assertEqual(distribution, ES_DISTRIBUTION)

    @httpretty.activate
    def test_check_instance_es_major_5(self):
        """Test whether the major version is correctly calculated for ElasticSearch 5.x"""

        body = """{
            "name" : "Amber Hunt",
            "cluster_name" : "jgbarah",
            "version" : {
                "number" : "5.0.0-alpha2",
                "build_hash" : "e3126df",
                "build_date" : "2016-04-26T12:08:58.960Z",
                "build_snapshot" : false,
                "lucene_version" : "6.0.0"
            },
            "tagline" : "You Know, for Search"
        }"""

        es_con = "http://es5.com"
        httpretty.register_uri(httpretty.GET,
                               es_con,
                               body=body,
                               status=200)

        major, distribution = ElasticSearch.check_instance(es_con, insecure=False)
        self.assertEqual(major, '5')
        self.assertEqual(distribution, ES_DISTRIBUTION)

    @httpretty.activate
    def test_check_instance_os_major_1(self):
        """Test whether the major version is correctly calculated for OpenSearch 1.x"""

        body = """{
            "name" : "AAHAA",
            "cluster_name" : "docker-cluster",
            "cluster_uuid" : "XABU6jucSkG48cd7ccUTxQ",
            "version" : {
                "distribution" : "opensearch",
                "number" : "1.2.4",
                "build_type" : "tar",
                "build_hash" : "e505b10357c03ae8d26d675172402f2f2144ef0f",
                "build_date" : "2022-01-14T03:38:06.881862Z",
                "build_snapshot" : false,
                "lucene_version" : "8.10.1",
                "minimum_wire_compatibility_version" : "6.8.0",
                "minimum_index_compatibility_version" : "6.0.0-beta1"
            },
            "tagline" : "The OpenSearch Project: https://opensearch.org/"
        }"""

        os_con = "http://os1.com"
        httpretty.register_uri(httpretty.GET,
                               os_con,
                               body=body,
                               status=200)

        major, distribution = ElasticSearch.check_instance(os_con, insecure=False)
        self.assertEqual(major, '1')
        self.assertEqual(distribution, OS_DISTRIBUTION)

    @httpretty.activate
    def test_check_instance_es_major_error(self):
        """Test whether an exception is thrown when the ElasticSearch version number is not retrieved"""

        body = """{
                "name" : "Amber Hunt",
                "cluster_name" : "jgbarah",
                "version" : {
                    "build_hash" : "e3126df",
                    "build_date" : "2016-04-26T12:08:58.960Z",
                    "build_snapshot" : false,
                    "lucene_version" : "6.0.0"
                },
                "tagline" : "You Know, for Search"
            }"""

        es_con = "http://es_err.com"
        httpretty.register_uri(httpretty.GET,
                               es_con,
                               body=body,
                               status=200)

        with self.assertRaises(ElasticError):
            _, _ = ElasticSearch.check_instance(es_con, insecure=False)

    @httpretty.activate
    def test_check_instance_not_reachable(self):
        """Test whether an exception is thrown when the ElasticSearch is not reachable"""

        es_con = "http://es_err.com"
        httpretty.register_uri(httpretty.GET,
                               es_con,
                               body={},
                               status=400)

        with self.assertRaises(ElasticError):
            _ = ElasticSearch.check_instance(es_con, insecure=False)

    def test_init_clean(self):
        """Test whether the index is recreated if `clean` is True"""

        elastic = ElasticSearch(self.es_con, self.target_index, GitOcean.mapping)
        r = elastic.requests.get(elastic.index_url)
        self.assertEqual(r.status_code, 200)

        with self.assertLogs(logger, level='INFO') as cm:
            elastic = ElasticSearch(self.es_con, self.target_index, GitOcean.mapping, clean=True)
            self.assertRegex(cm.output[0], 'INFO:grimoire_elk.elastic:Deleted and created index*')

        r = elastic.requests.get(elastic.index_url)
        self.assertEqual(r.status_code, 200)

    def test_init_aliases(self):
        """Test whether aliases are correctly set"""

        expected_aliases = {
            'A': {},
            'B': {}
        }

        elastic = ElasticSearch(self.es_con, self.target_index, GitOcean.mapping, aliases=["A", "B"])
        self.assertEqual(elastic.url, self.es_con)
        self.assertEqual(elastic.index, self.target_index)
        self.assertEqual(elastic.index_url, self.es_con + "/" + self.target_index)

        r = elastic.requests.get(elastic.index_url + '/_alias')
        aliases = r.json()[self.target_index]['aliases']

        self.assertDictEqual(aliases, expected_aliases)

    def test_init_duplicated_aliases(self):
        """Test whether duplicated aliases are ignored"""

        expected_aliases = {
            'A': {},
            'B': {}
        }

        elastic = ElasticSearch(self.es_con, self.target_index, GitOcean.mapping, aliases=["A", "B", "A"])
        self.assertEqual(elastic.url, self.es_con)
        self.assertEqual(elastic.index, self.target_index)
        self.assertEqual(elastic.index_url, self.es_con + "/" + self.target_index)

        r = elastic.requests.get(elastic.index_url + '/_alias')
        aliases = r.json()[self.target_index]['aliases']

        self.assertDictEqual(aliases, expected_aliases)

    def test_safe_index(self):
        """Test whether the index name is correctly defined"""

        expected_index = "good_index"
        index_name = ElasticSearch.safe_index("good_index")
        self.assertEqual(index_name, expected_index)

        expected_index = "goodindex"
        index_name = ElasticSearch.safe_index("goodindex")
        self.assertEqual(index_name, expected_index)

        expected_index = "bad_index"
        index_name = ElasticSearch.safe_index("bad/index")
        self.assertEqual(index_name, expected_index)

    def test_create_index(self):
        """Test whether an index is created"""

        expected_response = 200
        elastic = ElasticSearch(self.es_con, self.target_index, GitOcean.mapping)

        response = elastic.requests.get(elastic.index_url).status_code
        self.assertEqual(response, expected_response)

    def test_create_index_clean(self):
        """Test whether an index is created"""

        ElasticSearch(self.es_con, self.target_index, GitOcean.mapping)
        with self.assertLogs(logger, level='INFO') as cm:
            ElasticSearch(self.es_con, self.target_index, GitOcean.mapping, clean=True)
            self.assertRegex(cm.output[0], 'INFO:grimoire_elk.elastic:Deleted and created index*')

    @httpretty.activate
    def test_create_index_error(self):
        """Test whether an error is thrown when the index isn't created"""

        url = self.es_con + '/' + self.target_index
        httpretty.register_uri(httpretty.GET,
                               url,
                               body={},
                               status=400,
                               forcing_headers={
                                   "Content-Type": "application/json"
                               })
        httpretty.register_uri(httpretty.PUT,
                               url,
                               body={},
                               status=400,
                               forcing_headers={
                                   "Content-Type": "application/json"
                               })

        elastic = MockElasticSearch(self.es_con, self.target_index)
        with self.assertRaises(ElasticError):
            _ = elastic.create_index()

    def test_create_mappings(self):
        """Test whether a mapping is correctly created"""

        elastic = ElasticSearch(self.es_con, self.target_index)
        elastic.create_mappings(GitOcean.mapping.get_elastic_mappings(elastic.major))

        r = elastic.requests.get(elastic.index_url + '/_mapping')
        mapping = r.json()
        self.assertIsNotNone(mapping[self.target_index]['mappings'])

    @httpretty.activate
    def test_create_mappings_error(self):
        """Test whether an error is thrown when the mapping cannot be created"""

        elastic = MockElasticSearch(self.es_con, self.target_index)

        if not elastic.is_legacy():
            url = elastic.index_url + "/_mapping"
        else:
            url = elastic.index_url + "/items/_mapping"

        httpretty.register_uri(httpretty.PUT,
                               url,
                               body={},
                               status=400,
                               forcing_headers={
                                   "Content-Type": "application/json"
                               })

        with self.assertLogs(logger, level='ERROR') as cm:
            elastic.create_mappings(GitOcean.mapping.get_elastic_mappings(elastic.major))
            self.assertRegex(cm.output[0], 'ERROR:grimoire_elk.elastic:Error creating ES mappings*')

    @httpretty.activate
    def test_create_mappings_templates_error(self):
        """Test whether an error is thrown when the templates cannot be created"""

        elastic = MockElasticSearch(self.es_con, self.target_index)

        if not elastic.is_legacy():
            url = elastic.index_url + "/_mapping"
        else:
            url = elastic.index_url + "/items/_mapping"

        httpretty.register_uri(httpretty.PUT,
                               url,
                               body={},
                               status=400,
                               forcing_headers={
                                   "Content-Type": "application/json"
                               })

        with self.assertLogs(logger, level='ERROR') as cm:
            elastic.create_mappings(Mapping.get_elastic_mappings(elastic.major))
            self.assertRegex(cm.output[0], 'ERROR:grimoire_elk.elastic:Can\'t add mapping*')

    def test_all_es_aliases(self):
        """Test whether all aliases are correctly returned"""

        expected_aliases = ['A', 'B', 'C']
        elastic = ElasticSearch(self.es_con, self.target_index, GitOcean.mapping, aliases=['A', 'B', 'C'])
        aliases = elastic.all_es_aliases()

        for ea in expected_aliases:
            self.assertIn(ea, aliases)

    @httpretty.activate
    def test_all_es_aliases_error(self):
        """Test whether a warning message is logged when the aliases are not returned"""

        url = self.es_con + "/_aliases"
        httpretty.register_uri(httpretty.GET,
                               url,
                               body={},
                               status=400,
                               forcing_headers={
                                   "Content-Type": "application/json"
                               })

        elastic = MockElasticSearch(self.es_con, self.target_index)
        with self.assertLogs(logger, level='WARNING') as cm:
            _ = elastic.all_es_aliases()
            self.assertRegex(cm.output[0],
                             'WARNING:grimoire_elk.elastic:Something went wrong when retrieving aliases*')

    def test_list_aliases(self):
        """Test whether the aliases of a given index are correctly listed"""

        expected_aliases = {
            'A': {},
            'B': {},
            'C': {}
        }

        elastic = ElasticSearch(self.es_con, self.target_index, GitOcean.mapping, aliases=['A', 'B', 'C'])
        aliases = elastic.list_aliases()
        self.assertDictEqual(aliases, expected_aliases)

    @httpretty.activate
    def test_list_aliases_warning(self):
        """Test whether a warning message is logged when the aliases for a given index are not returned"""

        url = self.es_con + '/' + self.target_index + "/_alias"
        httpretty.register_uri(httpretty.GET,
                               url,
                               body={},
                               status=400,
                               forcing_headers={
                                   "Content-Type": "application/json"
                               })

        elastic = MockElasticSearch(self.es_con, self.target_index)
        with self.assertLogs(logger, level='WARNING') as cm:
            _ = elastic.list_aliases()
            self.assertRegex(cm.output[0],
                             'WARNING:grimoire_elk.elastic:Something went wrong when retrieving aliases*')

    def test_add_aliases(self):
        """Test whether an alias is added to a given index"""

        elastic = ElasticSearch(self.es_con, self.target_index, GitOcean.mapping, aliases=['A', 'B', 'C'])

        expected_aliases = {
            'A': {},
            'B': {},
            'C': {}
        }
        aliases = elastic.list_aliases()
        self.assertDictEqual(aliases, expected_aliases)

        expected_aliases = {
            'A': {},
            'B': {},
            'C': {},
            'D': {}
        }
        elastic.add_alias('D')
        aliases = elastic.list_aliases()
        self.assertDictEqual(aliases, expected_aliases)

    def test_add_aliases_duplicated(self):
        """Test whether an alias isn't added when already present in a given index"""

        elastic = ElasticSearch(self.es_con, self.target_index, GitOcean.mapping, aliases=['A', 'B', 'C'])

        expected_aliases = {
            'A': {},
            'B': {},
            'C': {}
        }
        aliases = elastic.list_aliases()
        self.assertDictEqual(aliases, expected_aliases)

        elastic.add_alias('C')
        aliases = elastic.list_aliases()
        self.assertDictEqual(aliases, expected_aliases)

    @httpretty.activate
    def test_add_alias_warning(self):
        """Test whether a warning message is logged when the aliases for a given index are not returned"""

        url = self.es_con + "/_aliases"
        httpretty.register_uri(httpretty.POST,
                               url,
                               body={},
                               status=400,
                               forcing_headers={
                                   "Content-Type": "application/json"
                               })

        elastic = MockElasticSearch(self.es_con, self.target_index, mock_list_alias=True)
        with self.assertLogs(logger, level='WARNING') as cm:
            _ = elastic.add_alias('A')
            self.assertRegex(cm.output[0],
                             'WARNING:grimoire_elk.elastic:Something went wrong when adding an alias*')

    def test_bulk_upload(self):
        """Test whether items are correctly uploaded to an index"""

        items = json.loads(read_file('data/git.json'))
        elastic = ElasticSearch(self.es_con, self.target_index, GitOcean.mapping)

        new_items = elastic.bulk_upload(items, field_id="uuid")
        self.assertEqual(new_items, 11)

    def test_bulk_upload_max_bulk(self):
        """Test whether items are correctly uploaded to an index"""

        items = json.loads(read_file('data/git.json'))
        elastic = ElasticSearch(self.es_con, self.target_index, GitOcean.mapping)
        elastic.max_items_bulk = 2

        new_items = elastic.bulk_upload(items, field_id="uuid")
        self.assertEqual(new_items, 11)

    def test_bulk_upload_no_items(self):
        """Test whether items are correctly uploaded to an index"""

        elastic = ElasticSearch(self.es_con, self.target_index, GitOcean.mapping)
        new_items = elastic.bulk_upload([], field_id="uuid")

        self.assertEqual(new_items, 0)

    def test_safe_put_bulk(self):
        """Test whether items are correctly stored to the index"""

        items = json.loads(read_file('data/git.json'))
        data_json = items[0]
        bulk_json = '{{"index" : {{"_id" : "{}" }} }}\n'.format(data_json['uuid'])
        bulk_json += json.dumps(data_json) + "\n"

        elastic = ElasticSearch(self.es_con, self.target_index, GitOcean.mapping)
        bulk_url = elastic.get_bulk_url()
        inserted_items = elastic.safe_put_bulk(bulk_url, bulk_json)

        self.assertEqual(inserted_items, 1)

    def test_safe_put_bulk_errors(self):
        """Test whether an error message is logged when an item isn't inserted"""

        items = json.loads(read_file('data/git.json'))
        data_json = items[0]
        data_json['origin'] = ''.join(random.choice(string.ascii_letters) for x in range(66000))
        bulk_json = '{{"index" : {{"_id" : "{}" }} }}\n'.format(data_json['uuid'])
        bulk_json += json.dumps(data_json) + "\n"

        elastic = ElasticSearch(self.es_con, self.target_index, GitOcean.mapping)
        bulk_url = elastic.get_bulk_url()

        with self.assertLogs(logger, level='ERROR') as cm:
            inserted_items = elastic.safe_put_bulk(bulk_url, bulk_json)
            self.assertRegex(cm.output[0], "ERROR:grimoire_elk.elastic:Failed to insert data to ES*")

        self.assertEqual(inserted_items, 0)

    def test_get_bulk_url(self):
        """Test that the bulk_url is correctly formed"""

        elastic = MockElasticSearch(self.es_con, self.target_index,
                                    major='7', distribution=ES_DISTRIBUTION)

        expected_url = elastic.url + '/' + elastic.index + '/_bulk'
        self.assertEqual(elastic.get_bulk_url(), expected_url)

        elastic = MockElasticSearch(self.es_con, self.target_index,
                                    major='6', distribution=ES_DISTRIBUTION)
        expected_url = elastic.url + '/' + elastic.index + '/items/_bulk'
        self.assertEqual(elastic.get_bulk_url(), expected_url)

        elastic = MockElasticSearch(self.es_con, self.target_index,
                                    major='1', distribution=OS_DISTRIBUTION)

        expected_url = elastic.url + '/' + elastic.index + '/_bulk'
        self.assertEqual(elastic.get_bulk_url(), expected_url)

    def test_get_mapping_url(self):
        """Test that the mapping_url is correctly formed"""

        elastic = MockElasticSearch(self.es_con, self.target_index,
                                    major='7', distribution=ES_DISTRIBUTION)

        expected_url = elastic.url + '/' + elastic.index + '/_mapping'
        self.assertEqual(elastic.get_mapping_url(), expected_url)

        elastic = MockElasticSearch(self.es_con, self.target_index,
                                    major='6', distribution=ES_DISTRIBUTION)
        expected_url = elastic.url + '/' + elastic.index + '/items/_mapping'
        self.assertEqual(elastic.get_mapping_url(_type='items'), expected_url)

        elastic = MockElasticSearch(self.es_con, self.target_index,
                                    major='1', distribution=OS_DISTRIBUTION)

        expected_url = elastic.url + '/' + elastic.index + '/_mapping'
        self.assertEqual(elastic.get_mapping_url(), expected_url)

    def test_all_properties(self):
        """Test whether all index properties are correctly returned"""

        expected_properties = {
            'data': {
                'properties': {
                    'message': {
                        'type': 'text'
                    },
                    'AuthorDate': {
                        'type': 'date',
                        'format': 'EEE MMM d HH:mm:ss yyyy Z||EEE MMM d HH:mm:ss yyyy||strict_date_optional_time||epoch_millis'
                    },
                    'CommitDate': {
                        'type': 'date',
                        'format': 'EEE MMM d HH:mm:ss yyyy Z||EEE MMM d HH:mm:ss yyyy||strict_date_optional_time||epoch_millis'
                    }
                }
            }
        }
        elastic = ElasticSearch(self.es_con, self.target_index, GitOcean.mapping)
        properties = elastic.all_properties()

        self.assertDictEqual(properties, expected_properties)

    def test_all_properties_no_properties(self):
        """Test whether no properties are returned if they don't exist in the mapping"""

        elastic = ElasticSearch(self.es_con, self.target_index, Mapping)
        properties = elastic.all_properties()

        self.assertDictEqual(properties, {})

    @httpretty.activate
    def test_all_properties_error(self):
        """Test whether an error message is logged when the properties aren't retrieved"""

        if not ElasticSearch.is_legacy_static(self.es_major, self.es_distribution):
            url = self.es_con + '/' + self.target_index + '/_mapping'
        else:
            url = self.es_con + '/' + self.target_index + '/items/_mapping'

        httpretty.register_uri(httpretty.GET,
                               url,
                               body={},
                               status=400,
                               forcing_headers={
                                   "Content-Type": "application/json"
                               })

        elastic = MockElasticSearch(self.es_con, self.target_index,
                                    major=self.es_major, distribution=self.es_distribution)
        with self.assertLogs(logger, level='ERROR') as cm:
            elastic.all_properties()
            self.assertRegex(cm.output[0], 'ERROR:grimoire_elk.elastic:Error all attributes*')

    def test_get_last_date(self):
        """Test whether the last date is correctly returned"""

        items = json.loads(read_file('data/git.json'))
        elastic = ElasticSearch(self.es_con, self.target_index, GitOcean.mapping)
        new_items = elastic.bulk_upload(items, field_id="uuid")
        self.assertEqual(new_items, 11)

        # no filter
        last_date = elastic.get_last_date('updated_on')
        self.assertEqual(last_date.isoformat(), '2019-10-01T18:05:52+00:00')

        # filter including all items
        fltr = {
            'name': 'origin',
            'value': '/tmp/perceval_mc84igfc/gittest'
        }
        last_date = elastic.get_last_date('updated_on', filters_=[fltr])
        self.assertEqual(last_date.isoformat(), '2014-02-12T06:11:12+00:00')

        # filter including a sub-set og items
        fltr = {
            'name': 'perceval_version',
            'value': '0.9.11'
        }
        last_date = elastic.get_last_date('updated_on', filters_=[fltr])
        self.assertEqual(last_date.isoformat(), '2014-02-12T06:09:04+00:00')

    def test_get_last_offset(self):
        """Test whether the last offset is correctly returned"""

        items = json.loads(read_file('data/kitsune.json'))
        elastic = ElasticSearch(self.es_con, self.target_index, KitsuneOcean.mapping)
        new_items = elastic.bulk_upload(items, field_id="uuid")
        self.assertEqual(new_items, 4)

        # no filter
        last_offset = elastic.get_last_offset('offset')
        self.assertEqual(last_offset, 3)

        # filter including all items
        fltr = {
            'name': 'origin',
            'value': 'http://example.com'
        }
        last_offset = elastic.get_last_offset('offset', filters_=[fltr])
        self.assertEqual(last_offset, 3)

        # filter including a sub-set og items
        fltr = {
            'name': 'perceval_version',
            'value': '0.9.11'
        }
        last_offset = elastic.get_last_offset('offset', filters_=[fltr])
        self.assertEqual(last_offset, 1)

    def test_get_last_item_field(self):
        """Test whether the date/offset of the last item is correctly returned"""

        items = json.loads(read_file('data/git.json'))
        elastic = ElasticSearch(self.es_con, self.target_index, GitOcean.mapping)
        new_items = elastic.bulk_upload(items, field_id="uuid")
        self.assertEqual(new_items, 11)

        # no filter
        last_date = elastic.get_last_item_field('updated_on')
        self.assertEqual(last_date.isoformat(), '2019-10-01T18:05:52+00:00')

        # None filter
        last_date = elastic.get_last_item_field('updated_on', filters_=None)
        self.assertEqual(last_date.isoformat(), '2019-10-01T18:05:52+00:00')

        # Multiple filters
        fltrs = [
            {
                'name': 'origin',
                'value': '/tmp/perceval_mc84igfc/gittest'
            },
            {
                'name': 'perceval_version',
                'value': '0.9.11'
            }
        ]
        last_date = elastic.get_last_item_field('updated_on', filters_=fltrs)
        self.assertEqual(last_date.isoformat(), '2014-02-12T06:09:04+00:00')

        # Handle None filter
        fltrs = [
            {
                'name': 'origin',
                'value': '/tmp/perceval_mc84igfc/gittest'
            },
            {
                'name': 'perceval_version',
                'value': '0.9.11'
            },
            None
        ]
        last_date = elastic.get_last_item_field('updated_on', filters_=fltrs)
        self.assertEqual(last_date.isoformat(), '2014-02-12T06:09:04+00:00')

    def test_get_last_item_field_handle_invalid_date_error(self):
        """Test whether long timestamps are properly handled"""

        items = json.loads(read_file('data/git.json'))
        items[-1]['updated_on'] = items[-1]['updated_on'] * 1000
        elastic = ElasticSearch(self.es_con, self.target_index, GitOcean.mapping)
        new_items = elastic.bulk_upload(items, field_id="uuid")
        self.assertEqual(new_items, 11)

        last_date = elastic.get_last_item_field('updated_on')
        self.assertEqual(last_date.isoformat(), '2019-10-01T18:05:53.024000+00:00')

    def test_delete_items(self):
        """Test whether items are correctly deleted"""

        items = json.loads(read_file('data/git.json'))
        for item in items:
            timestamp = unixtime_to_datetime(item['timestamp'])
            item['timestamp'] = timestamp.isoformat()

        elastic = ElasticSearch(self.es_con, self.target_index, GitOcean.mapping)
        new_items = elastic.bulk_upload(items, field_id="uuid")
        self.assertEqual(new_items, 11)

        url = self.es_con + '/' + self.target_index + '/_count'

        elastic.delete_items(retention_time=90000000, time_field='timestamp')
        left_items = elastic.requests.get(url).json()['count']
        self.assertEqual(left_items, 11)

        elastic.delete_items(retention_time=1, time_field='timestamp')
        left_items = elastic.requests.get(url).json()['count']
        self.assertEqual(left_items, 0)

    def test_delete_items_wrong_retention(self):
        """Test whether no items are deleted if retention isn't defined or negative"""

        items = json.loads(read_file('data/git.json'))
        for item in items:
            timestamp = unixtime_to_datetime(item['timestamp'])
            item['timestamp'] = timestamp.isoformat()

        elastic = ElasticSearch(self.es_con, self.target_index, GitOcean.mapping)
        new_items = elastic.bulk_upload(items, field_id="uuid")
        self.assertEqual(new_items, 11)

        url = self.es_con + '/' + self.target_index + '/_count'

        elastic.delete_items(retention_time=None, time_field='timestamp')
        left_items = elastic.requests.get(url).json()['count']
        self.assertEqual(left_items, 11)

        elastic.delete_items(retention_time=-1, time_field='timestamp')
        left_items = elastic.requests.get(url).json()['count']
        self.assertEqual(left_items, 11)

    def test_delete_items_error(self):
        """Test whether an error message is logged if the items aren't deleted"""

        items = json.loads(read_file('data/git.json'))

        elastic = ElasticSearch(self.es_con, self.target_index, GitOcean.mapping)
        new_items = elastic.bulk_upload(items, field_id="uuid")
        self.assertEqual(new_items, 11)

        with self.assertLogs(logger, level='ERROR') as cm:
            elastic.delete_items(retention_time=1, time_field='timestamp')
            self.assertRegex(cm.output[0], 'ERROR:grimoire_elk.elastic:\\[items retention\\] Error deleted items*')


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Nishchith Shetty <inishchith@gmail.com>
#
import logging
import unittest

from base import TestBaseBackend
from grimoire_elk.raw.graal import GraalOcean
from grimoire_elk.enriched.colic import logger


HEADER_JSON = {"Content-Type": "application/json"}


class TestCoLic(TestBaseBackend):
    """Test CoLic backend"""

    connector = "colic"
    ocean_index = "test_" + connector
    enrich_index = "test_" + connector + "_enrich"

    def test_has_identites(self):
        """Test value of has_identities method"""

        enrich_backend = self.connectors[self.connector][2]()
        self.assertFalse(enrich_backend.has_identities())

    def test_items_to_raw(self):
        """Test whether JSON items are properly inserted into ES"""

        result = self._test_items_to_raw()

        self.assertGreater(result['items'], 0)
        self.assertGreater(result['raw'], 0)
        self.assertGreaterEqual(result['items'], result['raw'])

    def test_raw_to_enrich(self):
        """Test whether the raw index is properly enriched"""

        result = self._test_raw_to_enrich()

        self.assertGreater(result['raw'], 0)
        self.assertGreater(result['enrich'], 0)
        self.assertGreaterEqual(result['enrich'], result['raw'])

        enrich_backend = self.connectors[self.connector][2]()

        item = self.items[0]
        eitem = enrich_backend.get_rich_items(item)[0]
        self.assertEqual(eitem['licenses'], ["gpl-3.0"])
        self.assertEqual(eitem['has_license'], 1)
        self.assertEqual(eitem['license_name'], ["GNU General Public License 3.0"])
        self.assertEqual(eitem['copyrights'], ["Copyright (c) 2007 Free Software Foundation, Inc. <http://fsf.org/>"])
        self.assertEqual(eitem['has_copyright'], 1)
        self.assertEqual(eitem['modules'], [])
        self.assertEqual(eitem['file_path'], "LICENSE")

        item = self.items[1]
        eitem = enrich_backend.get_rich_items(item)[0]
        self.assertEqual(eitem['licenses'], ["gpl-3.0-plus"])
        self.assertEqual(eitem['has_license'], 1)
        self.assertEqual(eitem['license_name'], ["GNU General Public License 3.0 or later"])
        self.assertEqual(eitem['copyrights'], ["Copyright (c) 2015-2018 Bitergia"])
        self.assertEqual(eitem['has_copyright'], 1)
        self.assertEqual(eitem['modules'], ["graal"])
        self.assertEqual(eitem['file_path'], "graal/codecomplexity.py")

        item = self.items[2]
        eitem = enrich_backend.get_rich_items(item)[0]
        self.assertEqual(eitem['licenses'], ["gpl-3.0-plus"])
        self.assertEqual(eitem['has_license'], 1)
        self.assertEqual(eitem['license_name'], ["GNU General Public License 3.0 or later"])
        self.assertEqual(eitem['copyrights'], ["Copyright (c) 2015-2018 Bitergia"])
        self.assertEqual(eitem['has_copyright'], 1)
        self.assertEqual(eitem['modules'], ["graal"])
        self.assertEqual(eitem['file_path'], "graal/codecomplexity.py")

        item = self.items[3]
        eitem = enrich_backend.get_rich_items(item)[0]
        self.assertEqual(eitem['licenses'], [])
        self.assertEqual(eitem['has_license'], 0)
        self.assertEqual(eitem['license_name'], [])
        self.assertEqual(eitem['copyrights'], [])
        self.assertEqual(eitem['has_copyright'], 0)
        self.assertEqual(eitem['modules'], [])
        self.assertEqual(eitem['file_path'], "README.md")

        item = self.items[4]
        eitem = enrich_backend.get_rich_items(item)[0]
        self.assertEqual(eitem['licenses'], ["GPL-3.0"])
        self.assertEqual(eitem['has_license'], 1)
        self.assertEqual(eitem['license_name'], ["GPL-3.0"])
        self.assertEqual(eitem['copyrights'], [])
        self.assertEqual(eitem['has_copyright'], 0)
        self.assertEqual(eitem['modules'], ["tests"])
        self.assertEqual(eitem['file_path'], "tests/test_colic.py")

    def test_colic_analysis_study(self):
        """ Test that the colic analysis study works correctly """

        study, ocean_backend, enrich_backend = self._test_study('enrich_colic_analysis')

        with self.assertLogs(logger, level='INFO') as cm:

            if study.__name__ == "enrich_colic_analysis":
                study(ocean_backend, enrich_backend)
                self.assertEqual(cm.output[0], 'INFO:grimoire_elk.enriched.colic:[colic] study enrich-colic-analysis '
                                               'start')
                self.assertEqual(cm.output[-1], 'INFO:grimoire_elk.enriched.colic:[colic] study enrich-colic-analysis '
                                                'end')

    def test_perceval_params(self):
        """Test the extraction of perceval params from an URL"""

        url = "https://github.com/grimoirelab/perceval"
        expected_params = [
            'https://github.com/grimoirelab/perceval'
        ]
        self.assertListEqual(GraalOcean.get_perceval_params_from_url(url), expected_params)

        url = "https://github.com/grimoirelab/perceval /tmp/perceval-repo"
        expected_params = [
            'https://github.com/grimoirelab/perceval'
        ]
        self.assertListEqual(GraalOcean.get_perceval_params_from_url(url), expected_params)

    def test_copy_raw_fields(self):
        """Test copied raw fields"""

        self._test_raw_to_enrich()
        enrich_backend = self.connectors[self.connector][2]()

        for item in self.items:
            eitem = enrich_backend.get_rich_items(item)[0]
            for attribute in enrich_backend.RAW_FIELDS_COPY:
                if attribute in item:
                    self.assertEqual(item[attribute], eitem[attribute])
                else:
                    self.assertIsNone(eitem[attribute])


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(message)s')
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("requests").setLevel(logging.WARNING)
    unittest.main(warnings='ignore')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Valerio Cosentino <valcos@bitergia.com>
#

import logging
import re

from grimoirelab_toolkit.datetime import (datetime_utcnow,
                                          str_to_datetime)

from .utils import get_time_diff_days

from .enrich import Enrich, metadata
from ..elastic_mapping import Mapping as BaseMapping


MAX_SIZE_BULK_ENRICHED_ITEMS = 200
GEOLOCATION_INDEX = '/github/'
GITHUB = 'https://github.com/'
ISSUE_TYPE = 'issue'
PULL_TYPE = 'pull_request'
COMMENT_TYPE = 'comment'
ISSUE_COMMENT_TYPE = 'issue_comment'
REVIEW_COMMENT_TYPE = 'review_comment'
REPOSITORY_TYPE = 'repository'

USER_NOT_AVAILABLE = {'organizations': []}
DELETED_USER_LOGIN = 'ghost'
DELETED_USER_NAME = 'Deleted user'

logger = logging.getLogger(__name__)


class Mapping(BaseMapping):

    @staticmethod
    def get_elastic_mappings(es_major):
        """Get Elasticsearch mapping.

        geopoints type is not created in dynamic mapping

        :param es_major: major version of Elasticsearch, as string
        :returns:        dictionary with a key, 'items', with the mapping
        """

        mapping = """
        {
            "properties": {
               "merge_author_geolocation": {
                   "type": "geo_point"
               },
               "assignee_geolocation": {
                   "type": "geo_point"
               },
               "issue_state": {
                   "type": "keyword"
               },
               "pull_state": {
                   "type": "keyword"
               },
               "user_geolocation": {
                   "type": "geo_point"
               },
               "title_analyzed": {
                    "type": "text",
                    "index": true
               },
               "body_analyzed": {
                    "type": "text",
                    "index": true
               },
               "id": {
                    "type": "keyword"
               }
            }
        }
        """

        return {"items": mapping}


class GitHubEnrich2(Enrich):

    mapping = Mapping

    comment_roles = ['user_data']
    issue_roles = ['assignee_data', 'user_data']
    pr_roles = ['merged_by_data', 'user_data']
    roles = ['assignee_data', 'merged_by_data', 'user_data']

    def __init__(self, db_sortinghat=None, json_projects_map=None,
                 db_user='', db_password='', db_host='', db_path=None,
                 db_port=None, db_ssl=False, db_verify_ssl=True, db_tenant=None):
        super().__init__(db_sortinghat=db_sortinghat, json_projects_map=json_projects_map,
                         db_user=db_user, db_password=db_password, db_host=db_host,
                         db_port=db_port, db_path=db_path, db_ssl=db_ssl, db_verify_ssl=db_verify_ssl,
                         db_tenant=db_tenant)

        self.studies = []
        self.studies.append(self.enrich_geolocation)
        self.studies.append(self.enrich_feelings)
        self.studies.append(self.enrich_extra_data)
        self.studies.append(self.enrich_demography)

    def set_elastic(self, elastic):
        self.elastic = elastic

    def get_sh_backend_name(self):
        """Retrieve the backend name for SortingHat identities."""

        return "github"

    def get_field_author(self):
        return "user_data"

    def get_field_date(self):
        """ Field with the date in the JSON enriched items """
        return "grimoire_creation_date"

    def get_identities(self, item):
        """Return the identities from an item"""

        category = item['category']
        item = item['data']
        if category == "issue":
            identity_types = ['user', 'assignee']
            comments_attrs = ['comments_data']
        elif category == "pull_request":
            identity_types = ['user', 'merged_by']
            comments_attrs = ['review_comments_data', 'reviews_data']
        else:
            identity_types = []
            comments_attrs = []

        for identity in identity_types:
            identity_attr = identity + "_data"
            if item[identity] and identity_attr in item:
                # In user_data we have the full user data
                user = self.get_sh_identity(item[identity_attr])
                if user:
                    yield user

        for comments_attr in comments_attrs:
            comments = item.get(comments_attr, [])
            for comment in comments:
                user = self.get_sh_identity(comment['user_data'])
                if user:
                    yield user

    def get_sh_identity(self, item, identity_field=None):
        identity = {}

        user = item  # by default a specific user dict is expected

        if isinstance(item, dict) and 'data' in item:
            user = item['data'][identity_field]
        elif identity_field:
            user = item[identity_field]

        if not user:
            return identity

        identity['email'] = user.get('email', None)
        identity['name'] = user.get('name', None)
        if not identity['name']:
            identity['name'] = user.get('login', None)
        identity['username'] = user.get('username', None)
        if not identity['username']:
            identity['username'] = user.get('login', None)

        return identity

    def get_project_repository(self, eitem):
        repo = eitem['origin']
        return repo

    def get_time_to_first_attention(self, item):
        """Get the first date at which a comment or reaction was made to the issue by someone
        other than the user who created the issue
        """
        dates = []
        deleted_user_login = {'login': DELETED_USER_LOGIN}

        for comment in item['comments_data']:
            # Add deleted (ghost) user
            if not comment['user']:
                comment['user'] = deleted_user_login

            # skip comments of the issue creator
            if item['user']['login'] == comment['user']['login']:
                continue

            dates.append(str_to_datetime(comment['created_at']))

        for reaction in item['reactions_data']:
            # Add deleted (ghost) user
            if not reaction['user']:
                reaction['user'] = deleted_user_login

            # skip reactions of the issue creator
            if item['user']['login'] == reaction['user']['login']:
                continue

            dates.append(str_to_datetime(reaction['created_at']))

        if dates:
            return min(dates)

        return None

    def get_time_to_merge_request_response(self, item):
        """Get the first date at which a review was made on the PR by someone
        other than the user who created the PR
        """
        review_dates = []
        for comment in item['review_comments_data']:
            # Add deleted (ghost) user
            if not comment['user']:
                comment['user'] = {'login': DELETED_USER_LOGIN}

            # skip comments of the pull request creator
            if item['user']['login'] == comment['user']['login']:
                continue

            review_dates.append(str_to_datetime(comment['created_at']))

        if review_dates:
            return min(review_dates)

        return None

    def get_field_unique_id(self):
        return "id"

    def add_gelk_metadata(self, eitem):
        eitem['metadata__gelk_version'] = self.gelk_version
        eitem['metadata__gelk_backend_name'] = self.__class__.__name__
        eitem['metadata__enriched_on'] = datetime_utcnow().isoformat()

    @metadata
    def get_rich_item(self, item):

        rich_item = {}
        if item['category'] == 'issue':
            rich_item = self.__get_rich_issue(item)
        elif item['category'] == 'pull_request':
            rich_item = self.__get_rich_pull(item)
        elif item['category'] == 'repository':
            rich_item = self.__get_rich_repo(item)
        else:
            logger.error("[github] rich item not defined for GitHub category {}".format(
                         item['category']))

        self.add_repository_labels(rich_item)
        self.add_metadata_filter_raw(rich_item)
        return rich_item

    def enrich_issue(self, item, eitem):
        eitems = []

        comments = item['data'].get('comments_data', [])
        if comments:
            rich_item_comments = self.get_rich_issue_comments(comments, eitem)
            eitems.extend(rich_item_comments)

        return eitems

    def get_rich_issue_comments(self, comments, eitem):
        ecomments = []

        for comment in comments:
            ecomment = {}

            self.copy_raw_fields(self.RAW_FIELDS_COPY, eitem, ecomment)

            # Copy data from the enriched issue
            ecomment['issue_labels'] = eitem['issue_labels']
            ecomment['issue_id'] = eitem['issue_id']
            ecomment['issue_id_in_repo'] = eitem['issue_id_in_repo']
            ecomment['issue_url'] = eitem['issue_url']
            ecomment['issue_title'] = eitem['issue_title']
            ecomment['issue_state'] = eitem['issue_state']
            ecomment['issue_created_at'] = eitem['issue_created_at']
            ecomment['issue_updated_at'] = eitem['issue_updated_at']
            ecomment['issue_closed_at'] = eitem['issue_closed_at']
            ecomment['issue_pull_request'] = eitem['issue_pull_request']
            ecomment['github_repo'] = eitem['github_repo']
            ecomment['repository'] = eitem['repository']
            ecomment['item_type'] = COMMENT_TYPE
            ecomment['sub_type'] = ISSUE_COMMENT_TYPE

            # Copy data from the raw comment
            ecomment['body'] = comment['body'][:self.KEYWORD_MAX_LENGTH]
            ecomment['body_analyzed'] = comment['body']
            ecomment['url'] = comment['html_url']

            # extract reactions and add it to enriched item
            ecomment.update(self.__get_reactions(comment))

            ecomment['comment_updated_at'] = comment['updated_at']

            # Add id info to allow to coexistence of items of different types in the same index
            ecomment['id'] = '{}_issue_comment_{}'.format(eitem['id'], comment['id'])
            ecomment.update(self.get_grimoire_fields(comment['updated_at'], ISSUE_COMMENT_TYPE))
            # due to backtrack compatibility, `is_github2_*` is replaced with `is_github_*`
            ecomment.pop('is_github2_{}'.format(ISSUE_COMMENT_TYPE))
            ecomment['is_github_{}'.format(ISSUE_COMMENT_TYPE)] = 1
            ecomment['is_github_comment'] = 1

            # Add user_login
            user_data = comment.get('user_data', None)
            if not user_data:
                user_data = {
                    'login': DELETED_USER_LOGIN,
                    'name': DELETED_USER_NAME
                }
                comment['user_data'] = user_data
            ecomment['user_login'] = user_data['login']

            if self.sortinghat:
                ecomment.update(self.get_item_sh(comment, self.comment_roles, 'updated_at'))

            if self.prjs_map:
                ecomment.update(self.get_item_project(ecomment))

            if 'project' in eitem:
                ecomment['project'] = eitem['project']

            self.add_repository_labels(ecomment)
            self.add_metadata_filter_raw(ecomment)
            self.add_gelk_metadata(ecomment)

            ecomments.append(ecomment)

        return ecomments

    def enrich_pulls(self, item, eitem):
        eitems = []

        comments = item['data'].get('review_comments_data', [])
        reviews = item['data'].get('reviews_data', [])
        if comments:
            rich_item_comments = self.get_rich_pull_reviews(comments, eitem)
            eitems.extend(rich_item_comments)
        if reviews:
            rich_item_reviews = self.get_rich_pull_reviews(reviews, eitem)
            eitems.extend(rich_item_reviews)

        return eitems

    def get_rich_pull_reviews(self, comments, eitem):
        ecomments = []

        for comment in comments:
            # If the comment comes from a review is "Approve" or "Change requests"
            # there is a "submitted_at" instead of "updated_at"
            if 'updated_at' not in comment:
                comment['updated_at'] = comment['submitted_at']

            ecomment = {}

            self.copy_raw_fields(self.RAW_FIELDS_COPY, eitem, ecomment)

            # Review state
            ecomment['review_state'] = comment.get('state', '')

            # Copy data from the enriched pull
            ecomment['pull_labels'] = eitem['pull_labels']
            ecomment['pull_id'] = eitem['pull_id']
            ecomment['pull_id_in_repo'] = eitem['pull_id_in_repo']
            ecomment['issue_id_in_repo'] = eitem['issue_id_in_repo']
            ecomment['issue_title'] = eitem['issue_title']
            ecomment['issue_url'] = eitem['issue_url']
            ecomment['pull_url'] = eitem['pull_url']
            ecomment['pull_state'] = eitem['pull_state']
            ecomment['pull_created_at'] = eitem['pull_created_at']
            ecomment['pull_updated_at'] = eitem['pull_updated_at']
            ecomment['pull_merged_at'] = eitem['pull_merged_at']
            ecomment['pull_closed_at'] = eitem['pull_closed_at']
            ecomment['pull_merged'] = eitem['pull_merged']
            ecomment['pull_state'] = eitem['pull_state']
            ecomment['github_repo'] = eitem['github_repo']
            ecomment['repository'] = eitem['repository']
            ecomment['item_type'] = COMMENT_TYPE
            ecomment['sub_type'] = REVIEW_COMMENT_TYPE

            # Copy data from the raw comment
            ecomment['body'] = comment['body'][:self.KEYWORD_MAX_LENGTH]
            ecomment['body_analyzed'] = comment['body']
            ecomment['url'] = comment['html_url']

            # extract reactions and add it to enriched item
            ecomment.update(self.__get_reactions(comment))

            ecomment['comment_updated_at'] = comment['updated_at']
            ecomment['comment_created_at'] = comment.get('created_at', comment['updated_at'])

            # Add id info to allow to coexistence of items of different types in the same index
            ecomment['id'] = '{}_review_comment_{}'.format(eitem['id'], comment['id'])
            ecomment.update(self.get_grimoire_fields(comment['updated_at'], REVIEW_COMMENT_TYPE))
            # due to backtrack compatibility, `is_github2_*` is replaced with `is_github_*`
            ecomment.pop('is_github2_{}'.format(REVIEW_COMMENT_TYPE))
            ecomment['is_github_{}'.format(REVIEW_COMMENT_TYPE)] = 1
            ecomment['is_github_comment'] = 1

            # Add user_login
            user_data = comment.get('user_data', None)
            if not user_data:
                user_data = {
                    'login': DELETED_USER_LOGIN,
                    'name': DELETED_USER_NAME
                }
                comment['user_data'] = user_data
            ecomment['user_login'] = user_data['login']

            if self.sortinghat:
                ecomment.update(self.get_item_sh(comment, self.comment_roles, 'updated_at'))

            if self.prjs_map:
                ecomment.update(self.get_item_project(ecomment))

            if 'project' in eitem:
                ecomment['project'] = eitem['project']

            self.add_repository_labels(ecomment)
            self.add_metadata_filter_raw(ecomment)
            self.add_gelk_metadata(ecomment)

            ecomments.append(ecomment)

        return ecomments

    def enrich_items(self, ocean_backend):
        items_to_enrich = []
        num_items = 0
        ins_items = 0

        for item in ocean_backend.fetch():
            eitems = []

            eitem = self.get_rich_item(item)
            items_to_enrich.append(eitem)
            if item['category'] == ISSUE_TYPE:
                eitems = self.enrich_issue(item, eitem)
            elif item['category'] == PULL_TYPE:
                eitems = self.enrich_pulls(item, eitem)

            items_to_enrich.extend(eitems)

            if len(items_to_enrich) < MAX_SIZE_BULK_ENRICHED_ITEMS:
                continue

            num_items += len(items_to_enrich)
            ins_items += self.elastic.bulk_upload(items_to_enrich, self.get_field_unique_id())
            items_to_enrich = []

        if len(items_to_enrich) > 0:
            num_items += len(items_to_enrich)
            ins_items += self.elastic.bulk_upload(items_to_enrich, self.get_field_unique_id())

        if num_items != ins_items:
            missing = num_items - ins_items
            logger.error("%s/%s missing items for GitHub", str(missing), str(num_items))
        else:
            logger.info("%s items inserted for GitHub", str(num_items))

        return num_items

    def __get_reactions(self, item):
        reactions = {}

        item_reactions = item.get('reactions', {})
        # remove reactions url
        item_reactions.pop('url', None)
        for reaction in item_reactions:
            if reaction == '-1':
                reaction_name = 'thumb_down'
            elif reaction == '+1':
                reaction_name = 'thumb_up'
            else:
                reaction_name = reaction

            reactions['reaction_{}'.format(reaction_name)] = item_reactions[reaction]

        return reactions

    def __get_rich_pull(self, item):
        rich_pr = {}

        self.copy_raw_fields(self.RAW_FIELDS_COPY, item, rich_pr)
        # The real data
        pull_request = item['data']

        rich_pr['time_to_close_days'] = \
            get_time_diff_days(pull_request['created_at'], pull_request['closed_at'])

        if pull_request['state'] != 'closed':
            rich_pr['time_open_days'] = \
                get_time_diff_days(pull_request['created_at'], datetime_utcnow().replace(tzinfo=None))
        else:
            rich_pr['time_open_days'] = rich_pr['time_to_close_days']

        rich_pr['user_login'] = pull_request['user']['login']

        user = pull_request.get('user_data', None)
        if user is not None and user:
            rich_pr['user_name'] = user['name']
            rich_pr['author_name'] = user['name']
            rich_pr["user_domain"] = self.get_email_domain(user['email']) if user['email'] else None
            rich_pr['user_org'] = user['company']
            rich_pr['user_location'] = user['location']
            rich_pr['user_geolocation'] = None
        else:
            rich_pr['user_name'] = None
            rich_pr["user_domain"] = None
            rich_pr['user_org'] = None
            rich_pr['user_location'] = None
            rich_pr['user_geolocation'] = None
            rich_pr['author_name'] = None

        merged_by = pull_request.get('merged_by_data', None)
        if merged_by and merged_by != USER_NOT_AVAILABLE:
            rich_pr['merge_author_login'] = merged_by['login']
            rich_pr['merge_author_name'] = merged_by['name']
            rich_pr["merge_author_domain"] = self.get_email_domain(merged_by['email']) if merged_by['email'] else None
            rich_pr['merge_author_org'] = merged_by['company']
            rich_pr['merge_author_location'] = merged_by['location']
            rich_pr['merge_author_geolocation'] = None
        else:
            rich_pr['merge_author_name'] = None
            rich_pr['merge_author_login'] = None
            rich_pr["merge_author_domain"] = None
            rich_pr['merge_author_org'] = None
            rich_pr['merge_author_location'] = None
            rich_pr['merge_author_geolocation'] = None

        rich_pr['id'] = pull_request['id']
        rich_pr['pull_id'] = pull_request['id']
        rich_pr['pull_id_in_repo'] = pull_request['html_url'].split("/")[-1]
        rich_pr['issue_id_in_repo'] = pull_request['html_url'].split("/")[-1]
        rich_pr['repository'] = self.get_project_repository(rich_pr)
        rich_pr['issue_title'] = pull_request['title']
        rich_pr['issue_title_analyzed'] = pull_request['title']
        rich_pr['pull_state'] = pull_request['state']
        rich_pr['pull_created_at'] = pull_request['created_at']
        rich_pr['pull_updated_at'] = pull_request['updated_at']
        rich_pr['pull_merged'] = pull_request['merged']
        rich_pr['pull_merged_at'] = pull_request['merged_at']
        rich_pr['pull_closed_at'] = pull_request['closed_at']
        rich_pr['url'] = pull_request['html_url']
        rich_pr['pull_url'] = pull_request['html_url']
        rich_pr['issue_url'] = pull_request['html_url']

        # extract reactions and add it to enriched item
        rich_pr.update(self.__get_reactions(pull_request))

        labels = []
        [labels.append(label['name']) for label in pull_request['labels'] if 'labels' in pull_request]
        rich_pr['pull_labels'] = labels

        rich_pr['item_type'] = PULL_TYPE

        rich_pr['github_repo'] = rich_pr['repository'].replace(GITHUB, '')
        rich_pr['github_repo'] = re.sub('.git$', '', rich_pr['github_repo'])

        # GMD code development metrics
        rich_pr['forks'] = pull_request['base']['repo']['forks_count']
        rich_pr['code_merge_duration'] = get_time_diff_days(pull_request['created_at'],
                                                            pull_request['merged_at'])
        rich_pr['num_review_comments'] = pull_request['review_comments']

        rich_pr['time_to_merge_request_response'] = None
        if pull_request['review_comments'] != 0:
            min_review_date = self.get_time_to_merge_request_response(pull_request)
            rich_pr['time_to_merge_request_response'] = \
                get_time_diff_days(str_to_datetime(pull_request['created_at']), min_review_date)

        if self.prjs_map:
            rich_pr.update(self.get_item_project(rich_pr))

        if 'project' in item:
            rich_pr['project'] = item['project']

        rich_pr.update(self.get_grimoire_fields(pull_request['created_at'], PULL_TYPE))
        # due to backtrack compatibility, `is_github2_*` is replaced with `is_github_*`
        rich_pr.pop('is_github2_{}'.format(PULL_TYPE))
        rich_pr['is_github_{}'.format(PULL_TYPE)] = 1

        if self.sortinghat:
            item[self.get_field_date()] = rich_pr[self.get_field_date()]
            rich_pr.update(self.get_item_sh(item, self.pr_roles))

        return rich_pr

    def __get_rich_issue(self, item):
        rich_issue = {}

        self.copy_raw_fields(self.RAW_FIELDS_COPY, item, rich_issue)
        # The real data
        issue = item['data']

        rich_issue['time_to_close_days'] = \
            get_time_diff_days(issue['created_at'], issue['closed_at'])

        if issue['state'] != 'closed':
            rich_issue['time_open_days'] = \
                get_time_diff_days(issue['created_at'], datetime_utcnow().replace(tzinfo=None))
        else:
            rich_issue['time_open_days'] = rich_issue['time_to_close_days']

        rich_issue['user_login'] = issue['user']['login']

        user = issue.get('user_data', None)
        if user is not None and user:
            rich_issue['user_name'] = user['name']
            rich_issue['author_name'] = user['name']
            rich_issue["user_domain"] = self.get_email_domain(user['email']) if user['email'] else None
            rich_issue['user_org'] = user['company']
            rich_issue['user_location'] = user['location']
            rich_issue['user_geolocation'] = None
        else:
            rich_issue['user_name'] = None
            rich_issue["user_domain"] = None
            rich_issue['user_org'] = None
            rich_issue['user_location'] = None
            rich_issue['user_geolocation'] = None
            rich_issue['author_name'] = None

        assignee = issue.get('assignee_data', None)
        if assignee and assignee != USER_NOT_AVAILABLE:
            rich_issue['assignee_login'] = assignee['login']
            rich_issue['assignee_name'] = assignee['name']
            rich_issue["assignee_domain"] = self.get_email_domain(assignee['email']) if assignee['email'] else None
            rich_issue['assignee_org'] = assignee['company']
            rich_issue['assignee_location'] = assignee['location']
            rich_issue['assignee_geolocation'] = None
        else:
            rich_issue['assignee_name'] = None
            rich_issue['assignee_login'] = None
            rich_issue["assignee_domain"] = None
            rich_issue['assignee_org'] = None
            rich_issue['assignee_location'] = None
            rich_issue['assignee_geolocation'] = None

        rich_issue['id'] = issue['id']
        rich_issue['issue_id'] = issue['id']
        rich_issue['issue_id_in_repo'] = issue['html_url'].split("/")[-1]
        rich_issue['repository'] = self.get_project_repository(rich_issue)
        rich_issue['issue_title'] = issue['title']
        rich_issue['issue_title_analyzed'] = issue['title']
        rich_issue['issue_state'] = issue['state']
        rich_issue['issue_created_at'] = issue['created_at']
        rich_issue['issue_updated_at'] = issue['updated_at']
        rich_issue['issue_closed_at'] = issue['closed_at']
        rich_issue['url'] = issue['html_url']
        rich_issue['issue_url'] = issue['html_url']

        # extract reactions and add it to enriched item
        rich_issue.update(self.__get_reactions(issue))

        labels = []
        [labels.append(label['name']) for label in issue['labels'] if 'labels' in issue]
        rich_issue['issue_labels'] = labels

        rich_issue['item_type'] = ISSUE_TYPE
        rich_issue['issue_pull_request'] = True
        if 'head' not in issue.keys() and 'pull_request' not in issue.keys():
            rich_issue['issue_pull_request'] = False

        rich_issue['github_repo'] = rich_issue['repository'].replace(GITHUB, '')
        rich_issue['github_repo'] = re.sub('.git$', '', rich_issue['github_repo'])

        if self.prjs_map:
            rich_issue.update(self.get_item_project(rich_issue))

        if 'project' in item:
            rich_issue['project'] = item['project']

        rich_issue['time_to_first_attention'] = None
        if issue['comments'] + issue['reactions']['total_count'] != 0:
            rich_issue['time_to_first_attention'] = \
                get_time_diff_days(str_to_datetime(issue['created_at']),
                                   self.get_time_to_first_attention(issue))

        rich_issue.update(self.get_grimoire_fields(issue['created_at'], ISSUE_TYPE))
        # due to backtrack compatibility, `is_github2_*` is replaced with `is_github_*`
        rich_issue.pop('is_github2_{}'.format(ISSUE_TYPE))
        rich_issue['is_github_{}'.format(ISSUE_TYPE)] = 1

        if self.sortinghat:
            item[self.get_field_date()] = rich_issue[self.get_field_date()]
            rich_issue.update(self.get_item_sh(item, self.issue_roles))

        return rich_issue

    def __get_rich_repo(self, item):
        rich_repo = {}

        self.copy_raw_fields(self.RAW_FIELDS_COPY, item, rich_repo)

        repo = item['data']

        rich_repo['id'] = str(repo['fetched_on'])
        rich_repo['forks_count'] = repo['forks_count']
        rich_repo['subscribers_count'] = repo['subscribers_count']
        rich_repo['stargazers_count'] = repo['stargazers_count']
        rich_repo['fetched_on'] = repo['fetched_on']
        rich_repo['url'] = repo['html_url']

        if self.prjs_map:
            rich_repo.update(self.get_item_project(rich_repo))

        rich_repo.update(self.get_grimoire_fields(item['metadata__updated_on'], REPOSITORY_TYPE))
        # due to backtrack compatibility, `is_github2_*` is replaced with `is_github_*`
        rich_repo.pop('is_github2_{}'.format(REPOSITORY_TYPE))
        rich_repo['is_github_{}'.format(REPOSITORY_TYPE)] = 1

        return rich_repo



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Valerio Cosentino <valcos@bitergia.com>
#
import logging
import unittest

from base import TestBaseBackend
from grimoire_elk.raw.graal import GraalOcean


HEADER_JSON = {"Content-Type": "application/json"}


class TestCoSmellsDocker(TestBaseBackend):
    """Test CoSmellsDocker backend"""

    connector = "dockersmells"
    ocean_index = "test_" + connector
    enrich_index = "test_" + connector + "_enrich"

    def test_has_identites(self):
        """Test value of has_identities method"""

        enrich_backend = self.connectors[self.connector][2]()
        self.assertFalse(enrich_backend.has_identities())

    def test_items_to_raw(self):
        """Test whether JSON items are properly inserted into ES"""

        result = self._test_items_to_raw()

        self.assertGreater(result['items'], 0)
        self.assertGreater(result['raw'], 0)
        self.assertGreaterEqual(result['items'], result['raw'])

    def test_raw_to_enrich(self):
        """Test whether the raw index is properly enriched"""

        result = self._test_raw_to_enrich()

        self.assertGreater(result['raw'], 0)
        self.assertGreater(result['enrich'], 0)
        self.assertGreaterEqual(result['enrich'], result['raw'])

        items = [item for item in self.enrich_backend.fetch()]

        item = items[0]
        self.assertEqual(item['file_path'], 'tests/Dockerfile')
        self.assertEqual(item['origin'], 'https://github.com/chaoss/grimoirelab')
        self.assertIn('smell', item)
        self.assertIn('smell_code', item)
        self.assertIn('smell_line', item)

        item = items[1]
        self.assertEqual(item['file_path'], 'tests/Dockerfile')
        self.assertEqual(item['origin'], 'https://github.com/chaoss/grimoirelab')
        self.assertIn('smell', item)
        self.assertIn('smell_code', item)
        self.assertIn('smell_line', item)

        item = items[2]
        self.assertEqual(item['file_path'], 'tests/Dockerfile')
        self.assertEqual(item['origin'], 'https://github.com/chaoss/grimoirelab')
        self.assertIn('smell', item)
        self.assertIn('smell_code', item)
        self.assertIn('smell_line', item)

        item = items[3]
        self.assertEqual(item['file_path'], 'tests/Dockerfile')
        self.assertEqual(item['origin'], 'https://github.com/chaoss/grimoirelab')
        self.assertIn('smell', item)
        self.assertIn('smell_code', item)
        self.assertIn('smell_line', item)

    def test_raw_to_enrich_projects(self):
        """Test enrich with Projects"""

        result = self._test_raw_to_enrich(projects=True)
        items = [item for item in self.enrich_backend.fetch()]

        item = items[0]
        self.assertEqual(item['file_path'], 'tests/Dockerfile')
        self.assertEqual(item['origin'], 'https://github.com/chaoss/grimoirelab')
        self.assertIn('smell', item)
        self.assertIn('smell_code', item)
        self.assertIn('smell_line', item)
        self.assertEqual(item['project'], 'Main')
        self.assertEqual(item['project_1'], 'Main')

        item = items[1]
        self.assertEqual(item['file_path'], 'tests/Dockerfile')
        self.assertEqual(item['origin'], 'https://github.com/chaoss/grimoirelab')
        self.assertIn('smell', item)
        self.assertIn('smell_code', item)
        self.assertIn('smell_line', item)
        self.assertEqual(item['project'], 'Main')
        self.assertEqual(item['project_1'], 'Main')

        item = items[2]
        self.assertEqual(item['file_path'], 'tests/Dockerfile')
        self.assertEqual(item['origin'], 'https://github.com/chaoss/grimoirelab')
        self.assertIn('smell', item)
        self.assertIn('smell_code', item)
        self.assertIn('smell_line', item)
        self.assertEqual(item['project'], 'Main')
        self.assertEqual(item['project_1'], 'Main')

        item = items[3]
        self.assertEqual(item['file_path'], 'tests/Dockerfile')
        self.assertEqual(item['origin'], 'https://github.com/chaoss/grimoirelab')
        self.assertIn('smell', item)
        self.assertIn('smell_code', item)
        self.assertIn('smell_line', item)
        self.assertEqual(item['project'], 'Main')
        self.assertEqual(item['project_1'], 'Main')

    def test_perceval_params(self):
        """Test the extraction of perceval params from an URL"""

        url = "https://github.com/grimoirelab/perceval"
        expected_params = [
            'https://github.com/grimoirelab/perceval'
        ]
        self.assertListEqual(GraalOcean.get_perceval_params_from_url(url), expected_params)

        url = "https://github.com/grimoirelab/perceval /tmp/perceval-repo"
        expected_params = [
            'https://github.com/grimoirelab/perceval'
        ]
        self.assertListEqual(GraalOcean.get_perceval_params_from_url(url), expected_params)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(message)s')
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("requests").setLevel(logging.WARNING)
    unittest.main(warnings='ignore')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Valerio Cosentino <valcos@bitergia.com>
#

import configparser
import logging
import json
import os
import unittest

import requests

from grimoire_elk.elastic import ElasticSearch
from grimoire_elk.elastic_items import (ElasticItems,
                                        logger)
from grimoirelab_toolkit.datetime import str_to_datetime
from grimoire_elk.raw.kitsune import KitsuneOcean
from grimoire_elk.raw.git import GitOcean
from grimoire_elk.raw.confluence import ConfluenceOcean
from grimoire_elk.raw.meetup import MeetupOcean
from perceval.backends.mozilla.kitsune import Kitsune
from perceval.backends.core.git import Git
from perceval.backends.core.confluence import Confluence
from perceval.backends.core.meetup import Meetup


CONFIG_FILE = 'tests.conf'


def read_file(filename, mode='r'):
    with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), filename), mode) as f:
        content = f.read()
    return content


class TestElasticItems(unittest.TestCase):
    """Unit tests for ElasticItems class"""

    target_index = "elastic_items_test"

    @classmethod
    def setUpClass(cls):
        cls.perceval_backend = Git('http://example.com', '/tmp/foo')
        cls.config = configparser.ConfigParser()
        cls.config.read(CONFIG_FILE)
        cls.es_con = dict(cls.config.items('ElasticSearch'))['url']

    def tearDown(self):
        target_index_url = self.es_con + "/" + self.target_index
        requests.delete(target_index_url, verify=False)

    def test_init(self):
        """Test whether attributes are initialized"""

        eitems = ElasticItems(self.perceval_backend)
        self.assertIsNone(eitems.from_date)
        self.assertIsNone(eitems.offset)
        self.assertIsInstance(eitems.requests, requests.Session)

        from_date = str_to_datetime("2019-01-01")
        eitems = ElasticItems(self.perceval_backend, from_date=from_date)
        self.assertEqual(eitems.from_date, from_date)
        self.assertIsNone(eitems.offset)
        self.assertIsInstance(eitems.requests, requests.Session)
        self.assertFalse(eitems.requests.verify)

        offset = 10
        eitems = ElasticItems(self.perceval_backend, offset=offset)
        self.assertIsNone(eitems.from_date)
        self.assertEqual(eitems.offset, offset)
        self.assertIsInstance(eitems.requests, requests.Session)
        self.assertFalse(eitems.requests.verify)

        insecure = False
        eitems = ElasticItems(self.perceval_backend, insecure=insecure)
        self.assertIsNone(eitems.from_date)
        self.assertIsNone(eitems.offset)
        self.assertIsInstance(eitems.requests, requests.Session)
        self.assertTrue(eitems.requests.verify)

    def test_get_repository_filter_raw(self):
        """Test whether the repository filter raw works properly"""

        expected_filter = {
            'name': 'origin',
            'value': 'http://example.com'
        }
        eitems = ElasticItems(self.perceval_backend)
        fltr = eitems.get_repository_filter_raw()
        self.assertDictEqual(fltr, expected_filter)

        expected_filter = {
            'name': 'tag',
            'value': 'test'
        }
        perceval_backend_meetup = Meetup('mygroup', 'aaaa', tag='test')
        eitems = MeetupOcean(perceval_backend_meetup)
        fltr = eitems.get_repository_filter_raw()
        self.assertDictEqual(fltr, expected_filter)

        expected_filter = {}
        perceval_backend_meetup = Meetup('mygroup', 'aaaa', tag='https://meetup.com/')
        eitems = MeetupOcean(perceval_backend_meetup)
        fltr = eitems.get_repository_filter_raw()
        self.assertDictEqual(fltr, expected_filter)

    def test_get_confluence_spaces(self):
        """Test whether the confluence spaces works properly"""

        # No backend
        expected_filter = {}
        eitems = ElasticItems(self.perceval_backend)
        fltr = eitems.get_confluence_spaces(['TEST'])
        self.assertDictEqual(fltr, expected_filter)

        # Confluence backend with spaces
        expected_filter = {
            "should": [
                {
                    "term": {"data._expandable.space": "/rest/api/space/TEST"}
                }
            ]
        }
        perceval_backend_confluence = Confluence('https://example.com', spaces=['TEST'])
        eitems = ConfluenceOcean(perceval_backend_confluence)
        fltr = eitems.get_confluence_spaces(perceval_backend_confluence.spaces)
        self.assertDictEqual(fltr, expected_filter)

        # Confluence backend without spaces
        expected_filter = {}
        perceval_backend_confluence = Confluence('https://example.com')
        eitems = ConfluenceOcean(perceval_backend_confluence)
        fltr = eitems.get_confluence_spaces(perceval_backend_confluence.spaces)
        self.assertDictEqual(fltr, expected_filter)

        # Use a different backend to check the result is {}
        expected_filter = {}
        perceval_backend_meetup = Meetup('mygroup', 'aaaa')
        eitems = MeetupOcean(perceval_backend_meetup)
        fltr = eitems.get_confluence_spaces(['TEST'])
        self.assertDictEqual(fltr, expected_filter)

    def test_get_field_date(self):
        """Test whether the field date is correctly returned"""

        eitems = ElasticItems(self.perceval_backend)
        self.assertEqual(eitems.get_field_date(), 'metadata__updated_on')

    def test_get_incremental_date(self):
        """Test whether the incremental date is correctly returned"""

        eitems = ElasticItems(self.perceval_backend)
        self.assertEqual(eitems.get_incremental_date(), 'metadata__timestamp')

    def test_set_projects_json_repo(self):
        """Test whether the projects json repo is correctly set"""

        expected_repo = "http://example.com --filter-raw=data.commit:51a3b654f"

        eitems = ElasticItems(self.perceval_backend)
        self.assertIsNone(eitems.projects_json_repo)
        eitems.set_projects_json_repo(expected_repo)
        self.assertEqual(eitems.projects_json_repo, expected_repo)

    def test_set_repo_labels(self):
        """Test whether the repo labels are correctly set"""

        expected_labels = ["A", "B", "C"]

        eitems = ElasticItems(self.perceval_backend)
        self.assertIsNone(eitems.repo_labels)
        eitems.set_repo_labels(expected_labels)
        self.assertListEqual(eitems.repo_labels, expected_labels)

    def test_set_repo_spaces(self):
        """Test whether the repo labels are correctly set"""

        expected_spaces = ["A", "B", "C"]

        eitems = ElasticItems(self.perceval_backend)
        self.assertIsNone(eitems.repo_spaces)
        eitems.set_repo_spaces(expected_spaces)
        self.assertListEqual(eitems.repo_spaces, expected_spaces)

    def test_extract_repo_tags(self):
        """Test whether the tags are correctly extracted from a URL repo"""

        eitems = ElasticItems(self.perceval_backend)
        # labels (by default)
        processed_repo, label_lst = eitems.extract_repo_tags("http://example.com --labels=[A, B, C]")

        self.assertEqual(processed_repo, "http://example.com")
        self.assertListEqual(label_lst, ['A', 'B', 'C'])

        # spaces
        processed_repo, space_lst = eitems.extract_repo_tags("http://example.com --spaces=[A, B, C]", tag="spaces")

        self.assertEqual(processed_repo, "http://example.com")
        self.assertListEqual(space_lst, ['A', 'B', 'C'])

        # spaces without tag='spaces'
        processed_repo, space_lst = eitems.extract_repo_tags("http://example.com --spaces=[A, B, C]")

        self.assertEqual(processed_repo, "http://example.com --spaces=[A, B, C]")
        self.assertListEqual(space_lst, [])

    def test_set_filter_raw(self):
        """Test whether the filter raw is properly set"""

        ei = ElasticItems(None)

        filter_raws = [
            "data.product:Firefox, for Android,data.component:Logins, Passwords and Form Fill",
            "data.product:Add-on SDK",
            "data.product:Add-on SDK,    data.component:Documentation",
            "data.product:Add-on SDK, data.component:General",
            "data.product:addons.mozilla.org Graveyard,       data.component:API",
            "data.product:addons.mozilla.org Graveyard,   data.component:Add-on Builder",
            "data.product:Firefox for Android,data.component:Build Config & IDE Support",
            "data.product:Firefox for Android,data.component:Logins, Passwords and Form Fill",
            "data.product:Mozilla Localizations,data.component:nb-NO / Norwegian Bokm\u00e5l",
            "data.product:addons.mozilla.org Graveyard,data.component:Add-on Validation"
        ]

        expected = [
            [
                {
                    "name": "data.product",
                    "value": "Firefox, for Android"
                },
                {
                    "name": "data.component",
                    "value": "Logins, Passwords and Form Fill"
                }
            ],
            [
                {
                    "name": "data.product",
                    "value": "Add-on SDK"
                }
            ],
            [
                {
                    "name": "data.product",
                    "value": "Add-on SDK"
                },
                {
                    "name": "data.component",
                    "value": "Documentation"
                }
            ],
            [
                {
                    "name": "data.product",
                    "value": "Add-on SDK"
                },
                {
                    "name": "data.component",
                    "value": "General"
                }
            ],
            [
                {
                    "name": "data.product",
                    "value": "addons.mozilla.org Graveyard"
                },
                {
                    "name": "data.component",
                    "value": "API"
                }
            ],
            [
                {
                    "name": "data.product",
                    "value": "addons.mozilla.org Graveyard"
                },
                {
                    "name": "data.component",
                    "value": "Add-on Builder"
                }
            ],
            [
                {
                    "name": "data.product",
                    "value": "Firefox for Android"
                },
                {
                    "name": "data.component",
                    "value": "Build Config & IDE Support"
                }
            ],
            [
                {
                    "name": "data.product",
                    "value": "Firefox for Android"
                },
                {
                    "name": "data.component",
                    "value": "Logins, Passwords and Form Fill"
                }
            ],
            [
                {
                    "name": "data.product",
                    "value": "Mozilla Localizations"
                },
                {
                    "name": "data.component",
                    "value": "nb-NO / Norwegian Bokm\u00e5l"
                }
            ],
            [
                {
                    "name": "data.product",
                    "value": "addons.mozilla.org Graveyard"
                },
                {
                    "name": "data.component",
                    "value": "Add-on Validation"
                }
            ]
        ]

        for i, filter_raw in enumerate(filter_raws):
            ei.set_filter_raw(filter_raw)

            self.assertDictEqual(ei.filter_raw_dict[0], expected[i][0])

            if len(ei.filter_raw_dict) > 1:
                self.assertDictEqual(ei.filter_raw_dict[1], expected[i][1])

    def test_get_connector_name(self):
        """Test whether the connector name is correctly returned"""

        eitems = ElasticItems(self.perceval_backend)
        self.assertIsNone(eitems.get_connector_name())

        eitems = GitOcean(self.perceval_backend)
        self.assertEqual(eitems.get_connector_name(), 'git')

    def test_set_cfg_section_name(self):
        """Test whether the section name is correctly set"""

        expected_section = "git"

        eitems = ElasticItems(self.perceval_backend)
        self.assertIsNone(eitems.cfg_section_name)
        eitems.set_cfg_section_name(expected_section)
        self.assertEqual(eitems.cfg_section_name, expected_section)

    def test_set_from_date(self):
        """Test whether the from date is correctly set"""

        expected_from_date = str_to_datetime("2019-01-01")

        eitems = ElasticItems(self.perceval_backend)
        self.assertIsNone(eitems.from_date)
        eitems.set_from_date(expected_from_date)
        self.assertEqual(eitems.from_date, expected_from_date)

    def test_fetch(self):
        """Test whether the fetch method properly works"""

        perceval_backend = Git('/tmp/perceval_mc84igfc/gittest', '/tmp/foo')
        elastic = ElasticSearch(self.es_con, self.target_index, GitOcean.mapping)

        # Load items
        items = json.loads(read_file('data/git.json'))
        ocean = GitOcean(perceval_backend)
        ocean.elastic = elastic
        ocean.feed_items(items)

        eitems = ElasticItems(perceval_backend)
        eitems.scroll_size = 2
        eitems.elastic = elastic

        items = [ei for ei in eitems.fetch()]
        self.assertEqual(len(items), 9)

    def test_fetch_from_date(self):
        """Test whether the fetch method with from_date properly works"""

        perceval_backend = Git('/tmp/perceval_mc84igfc/gittest', '/tmp/foo')
        elastic = ElasticSearch(self.es_con, self.target_index, GitOcean.mapping)

        # Load items
        items = json.loads(read_file('data/git.json'))
        ocean = GitOcean(perceval_backend)
        ocean.elastic = elastic
        ocean.feed_items(items)

        # Fetch total items
        eitems = ElasticItems(perceval_backend)
        eitems.elastic = elastic
        items = [ei for ei in eitems.fetch()]
        self.assertEqual(len(items), 9)

        # Fetch with from date
        from_date = str_to_datetime("2018-02-09T08:33:22.699+00:00")
        eitems = ElasticItems(perceval_backend, from_date=from_date)
        eitems.elastic = elastic
        items = [ei for ei in eitems.fetch()]
        self.assertEqual(len(items), 2)

    def test_fetch_from_offset(self):
        """Test whether the fetch method with offset properly works"""

        perceval_backend = Kitsune('http://example.com')
        elastic = ElasticSearch(self.es_con, self.target_index, KitsuneOcean.mapping)

        # Load items
        items = json.loads(read_file('data/kitsune.json'))
        ocean = KitsuneOcean(perceval_backend)
        ocean.elastic = elastic
        ocean.feed_items(items)

        # Fetch total items
        eitems = ElasticItems(perceval_backend)
        eitems.elastic = elastic
        items = [ei for ei in eitems.fetch()]
        self.assertEqual(len(items), 4)

        # Fetch with offset
        eitems = ElasticItems(perceval_backend, offset=2)
        eitems.elastic = elastic
        items = [ei for ei in eitems.fetch()]
        self.assertEqual(len(items), 2)

    def test_fetch_no_results(self):
        """Test whether a message is logged when no results are found"""

        perceval_backend = Git('/tmp/perceval_mc84igfc/gittest-not_found', '/tmp/foo')
        elastic = ElasticSearch(self.es_con, self.target_index, GitOcean.mapping)

        # Load items
        items = json.loads(read_file('data/git.json'))
        ocean = GitOcean(perceval_backend)
        ocean.elastic = elastic
        ocean.feed_items(items)

        eitems = ElasticItems(perceval_backend)
        eitems.elastic = elastic

        with self.assertLogs(logger, level='DEBUG') as cm:
            items = [ei for ei in eitems.fetch()]
            self.assertEqual(len(items), 0)
            self.assertRegex(cm.output[-2], 'DEBUG:grimoire_elk.elastic_items:No results found.*')
            self.assertRegex(cm.output[-1], 'DEBUG:grimoire_elk.elastic_items:Releasing scroll_id=*')

    def test_fetch_empty(self):
        """Test whether the fetch method returns an empty list when the index is empty"""

        eitems = ElasticItems(self.perceval_backend)
        items = [ei for ei in eitems.fetch()]
        self.assertEqual(items, [])

    def test_fetch_filter_raw(self):
        """Test whether the fetch with filter raw properly works"""

        perceval_backend = Git('/tmp/perceval_mc84igfc/gittest', '/tmp/foo')
        elastic = ElasticSearch(self.es_con, self.target_index, GitOcean.mapping)

        # Load items
        items = json.loads(read_file('data/git.json'))
        ocean = GitOcean(perceval_backend)
        ocean.elastic = elastic
        ocean.feed_items(items)

        eitems = ElasticItems(perceval_backend)
        eitems.set_filter_raw("data.commit:87783129c3f00d2c81a3a8e585eb86a47e39891a")
        eitems.elastic = elastic
        items = [ei for ei in eitems.fetch()]
        self.assertEqual(len(items), 1)

    def test_get_elastic_items(self):
        """Test whether the elastic method works properly"""

        perceval_backend = Git('/tmp/perceval_mc84igfc/gittest', '/tmp/foo')
        elastic = ElasticSearch(self.es_con, self.target_index, GitOcean.mapping)

        # Load items
        items = json.loads(read_file('data/git.json'))
        ocean = GitOcean(perceval_backend)
        ocean.elastic = elastic
        ocean.feed_items(items)

        eitems = ElasticItems(perceval_backend)
        eitems.elastic = elastic
        r_json = eitems.get_elastic_items()

        total = r_json['hits']['total']
        total = total['value'] if isinstance(total, dict) else total
        self.assertEqual(total, 9)

    def test_get_elastic_items_filter(self):
        """Test whether the elastic method works properly with filter"""

        perceval_backend = Git('/tmp/perceval_mc84igfc/gittest', '/tmp/foo')
        elastic = ElasticSearch(self.es_con, self.target_index, GitOcean.mapping)

        # Load items
        items = json.loads(read_file('data/git.json'))
        ocean = GitOcean(perceval_backend)
        ocean.elastic = elastic
        ocean.feed_items(items)

        filter = {
            "name": "uuid",
            "value": [
                "43f217b2f678a5691fdbc5c6c5302243e79e5a90",
                "00ee6902e34b309cd05706c26e3e195a62492f60"
            ]
        }

        eitems = ElasticItems(perceval_backend)
        eitems.elastic = elastic
        r_json = eitems.get_elastic_items(_filter=filter)
        hits = r_json['hits']['hits']

        self.assertEqual(len(hits), 2)
        self.assertEqual(hits[0]['_source']['uuid'], "43f217b2f678a5691fdbc5c6c5302243e79e5a90")
        self.assertEqual(hits[1]['_source']['uuid'], "00ee6902e34b309cd05706c26e3e195a62492f60")

    def test_get_elastic_items_error(self):
        """Test whether a message is logged if an error occurs when getting items from an index"""

        items = json.loads(read_file('data/git.json'))
        perceval_backend = Git('/tmp/perceval_mc84igfc/gittest', '/tmp/foo')
        elastic = ElasticSearch(self.es_con, self.target_index, GitOcean.mapping)
        elastic.bulk_upload(items, field_id="uuid")

        # Load items
        eitems = ElasticItems(perceval_backend)
        eitems.elastic = elastic

        with self.assertLogs(logger, level='DEBUG') as cm:
            r_json = eitems.get_elastic_items()
            self.assertIsNone(r_json)
            self.assertRegex(cm.output[-1], 'DEBUG:grimoire_elk.elastic_items:No results found from*')


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(message)s')
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Nitish Gupta <imnitish.ng@gmail.com>
#
import logging
import unittest

import requests
from base import TestBaseBackend
from grimoire_elk.raw.launchpad import LaunchpadOcean


class TestLaunchpad(TestBaseBackend):
    """Test Launchpad backend"""

    connector = "launchpad"
    ocean_index = "test_" + connector
    enrich_index = "test_" + connector + "_enrich"

    def test_has_identities(self):
        """Test value of has_identities method"""

        enrich_backend = self.connectors[self.connector][2]()
        self.assertTrue(enrich_backend.has_identities())

    def test_items_to_raw(self):
        """Test whether JSON items are properly inserted into ES"""

        result = self._test_items_to_raw()

        self.assertEqual(result['items'], 4)
        self.assertEqual(result['raw'], 4)

    def test_raw_to_enrich(self):
        """Test whether the raw index is properly enriched"""

        result = self._test_raw_to_enrich()
        self.assertEqual(result['raw'], 4)
        self.assertEqual(result['enrich'], 4)

        enrich_backend = self.connectors[self.connector][2]()

        item = self.items[0]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['offset'], None)
        self.assertEqual(eitem['origin'], 'https://launchpad.net/mydistribution')
        self.assertEqual(eitem['tag'], 'https://launchpad.net/mydistribution')
        self.assertEqual(eitem['uuid'], 'db64288a362e6c086d344bed242fd5914f966e29')
        self.assertIn('title', eitem)
        self.assertEqual(eitem['web_link'], 'https://bugs.launchpad.net/mydistribution/+source/mypackage/+bug/1')
        self.assertEqual(eitem['date_created'], '2017-08-21T14:28:00.336698+00:00')
        self.assertEqual(eitem['date_incomplete'], None)
        self.assertEqual(eitem['is_complete'], False)
        self.assertEqual(eitem['status'], 'Triaged')
        self.assertEqual(eitem['bug_target_name'], 'synaptic (Ubuntu)')
        self.assertEqual(eitem['importance'], 'High')
        self.assertEqual(eitem['date_triaged'], '2017-08-21T14:50:35.930044+00:00')
        self.assertEqual(eitem['date_left_new'], '2017-08-21T14:50:35.930044+00:00')
        self.assertEqual(eitem['time_to_last_update_days'], 0.0)
        self.assertEqual(eitem['reopened'], 0)
        self.assertEqual(eitem['time_to_fix_commit'], None)
        self.assertEqual(eitem['time_worked_on'], None)
        self.assertEqual(eitem['time_to_confirm'], 0.02)
        self.assertEqual(eitem['user_login'], 'user')
        self.assertEqual(eitem['user_name'], 'Cristian Aravena Romero')
        self.assertEqual(eitem['user_joined'], '2005-06-15T02:17:43.115113+00:00')
        self.assertEqual(eitem['user_karma'], 1841)
        self.assertEqual(eitem['user_time_zone'], 'America/Santiago')
        self.assertEqual(eitem['assignee_login'], 'user')
        self.assertEqual(eitem['assignee_name'], 'Cristian Aravena Romero')
        self.assertEqual(eitem['assignee_joined'], '2005-06-15T02:17:43.115113+00:00')
        self.assertEqual(eitem['assignee_karma'], 1841)
        self.assertEqual(eitem['assignee_time_zone'], 'America/Santiago')
        self.assertEqual(eitem['bug_name'], None)
        self.assertEqual(eitem['bug_id'], 1)
        self.assertEqual(eitem['latest_patch_uploaded'], None)
        self.assertEqual(eitem['security_related'], False)
        self.assertEqual(eitem['private'], False)
        self.assertEqual(eitem['users_affected_count'], 3)
        self.assertIn('description', eitem)
        self.assertEqual(eitem['tags'], [])
        self.assertEqual(eitem['date_last_updated'], '2011-10-17T21:12:59.354332+00:00')
        self.assertEqual(eitem['message_count'], 2)
        self.assertEqual(eitem['heat'], 22)
        self.assertEqual(eitem['time_to_first_attention'], None)
        self.assertEqual(eitem['activity_count'], 1)
        self.assertEqual(eitem['is_launchpad_issue'], 1)
        self.assertEqual(eitem['time_created_to_last_update_days'], 2041.07)

        item = self.items[1]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['offset'], None)
        self.assertEqual(eitem['origin'], 'https://launchpad.net/mydistribution')
        self.assertEqual(eitem['tag'], 'https://launchpad.net/mydistribution')
        self.assertEqual(eitem['uuid'], 'c99e2f572749ea280b7f05cd5ea78d39bcd17a00')
        self.assertIn('title', eitem)
        self.assertEqual(eitem['web_link'], 'https://bugs.launchpad.net/mydistribution/+source/mypackage/+bug/2')
        self.assertEqual(eitem['date_created'], '2006-03-16T19:25:37.186986+00:00')
        self.assertEqual(eitem['date_incomplete'], None)
        self.assertEqual(eitem['is_complete'], False)
        self.assertEqual(eitem['status'], 'Confirmed')
        self.assertEqual(eitem['bug_target_name'], 'synaptic (Ubuntu)')
        self.assertEqual(eitem['importance'], 'Medium')
        self.assertEqual(eitem['date_triaged'], None)
        self.assertEqual(eitem['date_left_new'], '2010-11-01T00:30:06.794525+00:00')
        self.assertEqual(eitem['time_to_last_update_days'], 4175.79)
        self.assertEqual(eitem['reopened'], 1)
        self.assertEqual(eitem['time_to_fix_commit'], None)
        self.assertEqual(eitem['time_worked_on'], None)
        self.assertEqual(eitem['time_to_confirm'], 1690.21)
        self.assertEqual(eitem['user_login'], 'user')
        self.assertEqual(eitem['user_name'], 'Cristian Aravena Romero')
        self.assertEqual(eitem['user_joined'], '2005-06-15T02:17:43.115113+00:00')
        self.assertEqual(eitem['user_karma'], 1841)
        self.assertEqual(eitem['user_time_zone'], 'America/Santiago')
        self.assertEqual(eitem['bug_name'], None)
        self.assertEqual(eitem['bug_id'], 2)
        self.assertEqual(eitem['latest_patch_uploaded'], None)
        self.assertEqual(eitem['security_related'], False)
        self.assertEqual(eitem['private'], False)
        self.assertEqual(eitem['users_affected_count'], 6)
        self.assertEqual(eitem['description'], 'Posting here what gnome says ')
        self.assertEqual(eitem['tags'], ['amd64', 'apport-bug', 'artful', 'wayland', 'wayland-session'])
        self.assertEqual(eitem['date_last_updated'], '2017-09-16T01:57:40.183918+00:00')
        self.assertEqual(eitem['message_count'], 14)
        self.assertEqual(eitem['heat'], 120)
        self.assertEqual(eitem['time_to_first_attention'], None)
        self.assertEqual(eitem['activity_count'], 1)
        self.assertEqual(eitem['is_launchpad_issue'], 1)
        self.assertEqual(eitem['time_created_to_last_update_days'], 25.48)

        item = self.items[2]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['offset'], None)
        self.assertEqual(eitem['origin'], 'https://launchpad.net/mydistribution')
        self.assertEqual(eitem['tag'], 'https://launchpad.net/mydistribution')
        self.assertEqual(eitem['uuid'], 'affb86ac0f679f9df3b4b345db330b2c3d7df49d')
        self.assertIn('title', eitem)
        self.assertEqual(eitem['web_link'], 'https://bugs.launchpad.net/mydistribution/+source/mypackage/+bug/3')
        self.assertEqual(eitem['date_created'], '2006-07-28T09:10:11.023558+00:00')
        self.assertEqual(eitem['date_incomplete'], None)
        self.assertEqual(eitem['is_complete'], False)
        self.assertEqual(eitem['status'], 'Confirmed')
        self.assertEqual(eitem['bug_target_name'], 'synaptic (Ubuntu)')
        self.assertEqual(eitem['importance'], 'Medium')
        self.assertEqual(eitem['date_triaged'], None)
        self.assertEqual(eitem['date_left_new'], None)
        self.assertEqual(eitem['reopened'], 1)
        self.assertEqual(eitem['time_to_fix_commit'], None)
        self.assertEqual(eitem['time_worked_on'], None)
        self.assertEqual(eitem['time_to_confirm'], 1681.92)
        self.assertEqual(eitem['user_login'], 'user')
        self.assertEqual(eitem['user_name'], 'Cristian Aravena Romero')
        self.assertEqual(eitem['user_joined'], '2005-06-15T02:17:43.115113+00:00')
        self.assertEqual(eitem['user_karma'], 1841)
        self.assertEqual(eitem['user_time_zone'], 'America/Santiago')
        self.assertEqual(eitem['bug_name'], None)
        self.assertEqual(eitem['bug_id'], 3)
        self.assertEqual(eitem['latest_patch_uploaded'], None)
        self.assertEqual(eitem['security_related'], False)
        self.assertEqual(eitem['private'], False)
        self.assertEqual(eitem['users_affected_count'], 3)
        self.assertEqual(eitem['tags'], [])
        self.assertEqual(eitem['date_last_updated'], '2011-10-17T21:12:59.354332+00:00')
        self.assertEqual(eitem['message_count'], 13)
        self.assertEqual(eitem['heat'], 22)
        self.assertEqual(eitem['time_to_first_attention'], None)
        self.assertEqual(eitem['activity_count'], 0)
        self.assertEqual(eitem['is_launchpad_issue'], 1)
        self.assertEqual(eitem['time_created_to_last_update_days'], 2041.07)

        item = self.items[3]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['offset'], None)
        self.assertEqual(eitem['origin'], 'https://launchpad.net/ubuntu/+source/python-defaults')
        self.assertEqual(eitem['tag'], 'https://launchpad.net/ubuntu/+source/python-defaults')
        self.assertEqual(eitem['uuid'], '252bb297c901a660f8a9752997840a6882bbd2f2')
        self.assertIn('title', eitem)
        self.assertEqual(eitem['web_link'], 'https://bugs.launchpad.net/ubuntu/+source/python-defaults/+bug/9560')
        self.assertEqual(eitem['date_created'], '2004-10-28T20:45:27+00:00')
        self.assertEqual(eitem['date_incomplete'], None)
        self.assertEqual(eitem['is_complete'], True)
        self.assertEqual(eitem['status'], 'Fix Released')
        self.assertEqual(eitem['bug_target_name'], 'python-defaults (Ubuntu)')
        self.assertEqual(eitem['importance'], 'Medium')
        self.assertEqual(eitem['date_triaged'], None)
        self.assertEqual(eitem['date_left_new'], None)
        self.assertEqual(eitem['time_to_close_days'], 0.0)
        self.assertEqual(eitem['reopened'], 0)
        self.assertEqual(eitem['time_to_fix_commit'], None)
        self.assertEqual(eitem['time_worked_on'], None)
        self.assertEqual(eitem['time_to_confirm'], None)
        self.assertEqual(eitem['user_login'], 'debzilla')
        self.assertEqual(eitem['user_name'], 'Debian Bug Importer')
        self.assertEqual(eitem['user_joined'], '2006-01-13T12:51:52.192089+00:00')
        self.assertEqual(eitem['user_karma'], 0)
        self.assertEqual(eitem['user_time_zone'], 'UTC')
        self.assertEqual(eitem['assignee_login'], 'doko')
        self.assertEqual(eitem['assignee_name'], 'Matthias Klose')
        self.assertEqual(eitem['assignee_joined'], '2005-06-15T02:17:43.115113+00:00')
        self.assertEqual(eitem['assignee_karma'], 261688)
        self.assertEqual(eitem['assignee_time_zone'], 'Europe/Berlin')
        self.assertEqual(eitem['bug_name'], None)
        self.assertEqual(eitem['bug_id'], 9560)
        self.assertEqual(eitem['latest_patch_uploaded'], None)
        self.assertEqual(eitem['security_related'], False)
        self.assertEqual(eitem['private'], False)
        self.assertEqual(eitem['users_affected_count'], 0)
        self.assertIn('description', eitem)
        self.assertEqual(eitem['tags'], [])
        self.assertEqual(eitem['date_last_updated'], '2004-10-28T20:45:27+00:00')
        self.assertEqual(eitem['message_count'], 2)
        self.assertEqual(eitem['heat'], 6)
        self.assertEqual(eitem['time_to_first_attention'], 13.12)
        self.assertEqual(eitem['activity_count'], 0)
        self.assertEqual(eitem['is_launchpad_issue'], 1)
        self.assertEqual(eitem['time_created_to_last_update_days'], 0.0)

    def test_raw_to_enrich_sorting_hat(self):
        """Test enrich with SortingHat"""

        result = self._test_raw_to_enrich(sortinghat=True)
        self.assertEqual(result['raw'], 4)
        self.assertEqual(result['enrich'], 4)

        enrich_backend = self.connectors[self.connector][2]()
        enrich_backend.sortinghat = True

        url = self.es_con + "/" + self.enrich_index + "/_search"
        response = enrich_backend.requests.get(url, verify=False).json()
        for hit in response['hits']['hits']:
            source = hit['_source']
            if 'author_uuid' in source:
                self.assertIn('author_domain', source)
                self.assertIn('author_gender', source)
                self.assertIn('author_gender_acc', source)
                self.assertIn('author_org_name', source)
                self.assertIn('author_bot', source)
                self.assertIn('author_name', source)
                self.assertIn('author_multi_org_names', source)
                self.assertIn('owner_data_gender', source)
                self.assertIn('owner_data_gender_acc', source)
                self.assertIn('owner_data_multi_org_names', source)
                self.assertIn('owner_data_name', source)
                self.assertIn('owner_data_user_name', source)
                self.assertIn('owner_data_uuid', source)

    def test_raw_to_enrich_projects(self):
        """Test enrich with Projects"""

        result = self._test_raw_to_enrich(projects=True)
        self.assertEqual(result['raw'], 4)
        self.assertEqual(result['enrich'], 4)

        res = requests.get(self.es_con + "/" + self.enrich_index + "/_search", verify=False)
        for eitem in res.json()['hits']['hits']:
            self.assertEqual(eitem['_source']['project'], "Main")

    def test_perceval_params(self):
        """Test the extraction of perceval params from an URL"""

        url1 = "https://launchpad.net/ubuntu/+source/python-defaults"
        url2 = "https://launchpad.net/launchpad"
        expected_params = [
            ['ubuntu', '--package', 'python-defaults'],
            ['launchpad']
        ]
        self.assertListEqual(LaunchpadOcean.get_perceval_params_from_url(url1), expected_params[0])
        self.assertListEqual(LaunchpadOcean.get_perceval_params_from_url(url2), expected_params[1])

    def test_copy_raw_fields(self):
        """Test copied raw fields"""

        self._test_raw_to_enrich()
        enrich_backend = self.connectors[self.connector][2]()

        for item in self.items:
            eitem = enrich_backend.get_rich_item(item)
            for attribute in enrich_backend.RAW_FIELDS_COPY:
                if attribute in item:
                    self.assertEqual(item[attribute], eitem[attribute])
                else:
                    self.assertIsNone(eitem[attribute])

    def test_refresh_identities(self):
        """Test refresh identities"""

        result = self._test_refresh_identities()
        # ... ?


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(message)s')
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("requests").setLevel(logging.WARNING)
    unittest.main(warnings='ignore')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Valerio Cosentino <valcos@bitergia.com>
#

"""GoogleHits Ocean feeder"""

from .elastic import ElasticOcean


class GoogleHitsOcean(ElasticOcean):
    """GoogleHits Ocean feeder"""

    pass



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Valerio Cosentino <valcos@bitergia.com>
#

import logging

from .enrich import Enrich, metadata


logger = logging.getLogger(__name__)


class GoogleHitsEnrich(Enrich):

    def get_identities(self, item):
        """ Return the identities from an item """
        identities = []

        return identities

    def has_identities(self):
        """ Return whether the enriched items contains identities """

        return False

    @metadata
    def get_rich_item(self, item):
        eitem = {}

        self.copy_raw_fields(self.RAW_FIELDS_COPY, item, eitem)
        # The real data
        entry = item['data']

        for e in entry.keys():
            eitem[e] = entry[e]

        if self.prjs_map:
            eitem.update(self.get_item_project(eitem))

        eitem.update(self.get_grimoire_fields(item["metadata__updated_on"], "hits"))

        self.add_repository_labels(eitem)
        self.add_metadata_filter_raw(eitem)
        return eitem



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

import json
import logging
import sys
from time import time

import requests

from grimoirelab_toolkit.datetime import (str_to_datetime,
                                          unixtime_to_datetime,
                                          InvalidDateError)

from grimoire_elk.errors import ELKError, ElasticError
from grimoire_elk.enriched.utils import (grimoire_con,
                                         get_diff_current_date,
                                         anonymize_url)

logger = logging.getLogger(__name__)

HEADER_JSON = {"Content-Type": "application/json"}


class ElasticSearch(object):

    max_items_bulk = 1000
    max_items_clause = 1000  # max items in search clause (refresh identities)

    def __init__(self, url, index, mappings=None, clean=False,
                 insecure=True, analyzers=None, aliases=None):
        """Class to handle the operations with the ElasticSearch database, such as
        creating indexes, mappings, setting up aliases and uploading documents.

        :param url: ES url
        :param index: index name
        :param mappings: an instance of the Mapping class
        :param clean: if True, deletes an existing index and create it again
        :param insecure: support https with invalid certificates
        :param analyzers: analyzers for ElasticSearch
        :param aliases: list of aliases, defined as strings, to be added to the index
        """
        # Get major version of Elasticsearch instance
        major, distribution = self.check_instance(url, insecure)
        self.major = major
        self.distribution = distribution
        logger.debug("Found version of {} instance at {}: {}.".format(
                     self.distribution, anonymize_url(url), self.major))

        self.url = url

        # Valid index for elastic
        self.index = self.safe_index(index)
        self.aliases = aliases

        self.index_url = self.url + "/" + self.index
        self.wait_bulk_seconds = 2  # time to wait to complete a bulk operation

        self.requests = grimoire_con(insecure)

        analyzer_settings = None

        if analyzers:
            analyzers_dict = analyzers.get_elastic_analyzers(es_major=self.major)
            analyzer_settings = analyzers_dict['items']

        self.create_index(analyzer_settings, clean)

        if analyzers:
            self.update_analyzers(analyzer_settings)
        if mappings:
            map_dict = mappings.get_elastic_mappings(es_major=self.major)
            self.create_mappings(map_dict)

        if aliases:
            for alias in aliases:
                if self.alias_in_use(alias):
                    logger.debug("Alias {} won't be set on {}, it already exists on {}".format(
                                 alias, anonymize_url(self.index_url), anonymize_url(self.url)))
                    continue

                self.add_alias(alias)

    @classmethod
    def safe_index(cls, unique_id):
        """Return a valid elastic index generated from unique_id

        :param unique_id: index name
        """

        index = unique_id
        if unique_id:
            index = unique_id.replace("/", "_").lower()
        return index

    @staticmethod
    def check_instance(url, insecure):
        """Checks if there is an instance of ElasticSearch/OpenSearch in url.

        Actually, it checks if GET on the url returns a JSON document
        with a field tagline "You know, for search",
        and a field version.number.

        :value      url: url of the instance to check
        :value insecure: don't verify ssl connection (boolean)

        :returns:        major version, as str and the distribution name.
        """
        res = grimoire_con(insecure).get(url)
        if res.status_code != 200:
            msg = "Got {} from url {}".format(res.status_code, url)
            logger.error(msg)
            raise ElasticError(cause=msg)
        else:
            try:
                version_str = res.json()['version']['number']
                version_major = version_str.split('.')[0]
                distribution = res.json()['version'].get('distribution', 'elasticsearch')
                return version_major, distribution
            except Exception:
                msg = "Could not read proper welcome message from url {}, {}".format(
                    anonymize_url(url),
                    res.text
                )
                logger.error(msg)
                raise ElasticError(cause=msg)

    def create_index(self, analyzers=None, clean=False):
        """Create an index. If clean is `True`, the target index will be deleted and recreated.

        :param analyzers: set index analyzers
        :param clean: if True, the index is deleted and recreated
        """
        res = self.requests.get(self.index_url)

        headers = {"Content-Type": "application/json"}
        if res.status_code != 200:
            # Index does no exists
            res = self.requests.put(self.index_url, data=analyzers,
                                    headers=headers)
            if res.status_code != 200:
                msg = "Can't create index {} ({})".format(anonymize_url(self.index_url), res.status_code)
                logger.error(msg)
                raise ElasticError(cause=msg)
            else:
                logger.info("Created index {}".format(anonymize_url(self.index_url)))
        else:
            if clean:
                res = self.requests.delete(self.index_url)
                res.raise_for_status()
                res = self.requests.put(self.index_url, data=analyzers,
                                        headers=headers)
                res.raise_for_status()
                logger.info("Deleted and created index {}".format(anonymize_url(self.index_url)))

    def safe_put_bulk(self, url, bulk_json):
        """Bulk items to a target index `url`. In case of UnicodeEncodeError,
        the bulk is encoded with iso-8859-1.

        :param url: target index where to bulk the items
        :param bulk_json: str representation of the items to upload
        """
        headers = {"Content-Type": "application/x-ndjson"}

        try:
            res = self.requests.put(url + '?refresh=true', data=bulk_json, headers=headers)
            res.raise_for_status()
        except UnicodeEncodeError:
            # Related to body.encode('iso-8859-1'). mbox data
            logger.warning("Encondig error ... converting bulk to iso-8859-1")
            bulk_json = bulk_json.encode('iso-8859-1', 'ignore')
            res = self.requests.put(url, data=bulk_json, headers=headers)
            res.raise_for_status()

        result = res.json()
        failed_items = []
        error = ""
        if result['errors']:
            # Due to multiple errors that may be thrown when inserting bulk data, only the first error is returned
            failed_items = [item['index'] for item in result['items'] if 'error' in item['index']]
            error = str(failed_items[0]['error'])

            logger.error("Failed to insert data to ES: {}, {}".format(error, anonymize_url(url)))

        inserted_items = len(result['items']) - len(failed_items)

        # The exception is currently not thrown to avoid stopping ocean uploading processes
        try:
            if failed_items:
                raise ELKError(cause=error)
        except ELKError:
            pass

        logger.debug("{} items uploaded to ES ({})".format(inserted_items, anonymize_url(url)))
        return inserted_items

    def all_es_aliases(self):
        """List all aliases used in ES"""

        r = self.requests.get(self.url + "/_aliases", headers=HEADER_JSON, verify=False)
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as ex:
            logger.warning("Something went wrong when retrieving aliases on {}, {}".format(
                           anonymize_url(self.index_url), ex))
            return

        aliases = []
        for index in r.json().keys():
            aliases.extend(list(r.json()[index]['aliases'].keys()))

        aliases = list(set(aliases))
        return aliases

    def list_aliases(self):
        """List aliases linked to the index"""

        # check alias doesn't exist
        r = self.requests.get(self.index_url + "/_alias", headers=HEADER_JSON, verify=False)
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as ex:
            logger.warning("Something went wrong when retrieving aliases on {}, {}".format(
                           anonymize_url(self.index_url), ex))
            return

        aliases = r.json()[self.index]['aliases']
        return aliases

    def alias_in_use(self, alias):
        """Check that an alias is already used in the ElasticSearch database

        :param alias: target alias
        :return: bool
        """
        aliases = self.all_es_aliases()
        return alias in aliases

    def add_alias(self, alias):
        """Add an alias to the index set in the elastic obj

        :param alias: alias to add

        :returns: None
        """
        aliases = self.list_aliases()
        alias_dict = alias
        if isinstance(alias, str):
            alias_dict = {
                "alias": alias
            }

        if aliases and alias_dict['alias'] in aliases:
            logger.debug("Alias {} already exists on {}.".format(
                alias_dict['alias'],
                anonymize_url(self.index_url)
            ))
            return

        # add alias
        alias_dict['index'] = self.index
        alias_action = {
            "actions": [
                {
                    "add": alias_dict
                }
            ]
        }

        r = self.requests.post(self.url + "/_aliases", headers=HEADER_JSON, verify=False, data=json.dumps(alias_action))
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as ex:
            logger.warning("Something went wrong when adding an alias on {}. Alias not set. {}".format(
                           anonymize_url(self.index_url), ex))
            return

        logger.info("Alias {} created on {}.".format(alias, anonymize_url(self.index_url)))

    def get_bulk_url(self):
        """Get the bulk URL endpoint"""

        if not self.is_legacy():
            bulk_url = self.index_url + '/_bulk'
        else:
            bulk_url = self.index_url + '/items/_bulk'

        return bulk_url

    def get_mapping_url(self, _type=None):
        """Get the mapping URL endpoint

        :param _type: type of the mapping. In case of ES >= 7, it is None
        """
        if not self.is_legacy():
            mapping_url = self.index_url + "/_mapping"
        else:
            mapping_url = self.index_url + "/" + _type + "/_mapping"

        return mapping_url

    def bulk_upload(self, items, field_id):
        """Upload in controlled packs items to ES using bulk API

        :param items: list of items to be uploaded
        :param field_id: unique ID attribute used to differentiate the items
        """
        current = 0
        new_items = 0  # total items added with bulk
        bulk_json = ""

        if not items:
            return new_items

        url = self.get_bulk_url()

        logger.debug("Adding items to {} (in {} packs)".format(anonymize_url(url), self.max_items_bulk))
        task_init = time()

        for item in items:
            if current >= self.max_items_bulk:
                task_init = time()
                new_items += self.safe_put_bulk(url, bulk_json)
                current = 0
                json_size = sys.getsizeof(bulk_json) / (1024 * 1024)
                logger.debug("bulk packet sent ({:.2f} sec, {} total, {:.2f} MB)".format(
                             time() - task_init, new_items, json_size))
                bulk_json = ""
            data_json = json.dumps(item)
            bulk_json += '{{"index" : {{"_id" : "{}" }} }}\n'.format(item[field_id])
            bulk_json += data_json + "\n"  # Bulk document
            current += 1

        if current > 0:
            new_items += self.safe_put_bulk(url, bulk_json)
            json_size = sys.getsizeof(bulk_json) / (1024 * 1024)
            logger.debug("bulk packet sent ({:.2f} sec prev, {} total, {:.2f} MB)".format(
                         time() - task_init, new_items, json_size))

        return new_items

    def update_analyzers(self, analyzers):
        """Update the settings with the analyzer for a given index.
        To update the settings we have to:
        1. Close the index
        2. Update the settings
        3. Open the index.

        :param analyzers: elastic_analyzer.Analyzer object
        """
        if analyzers == '{}':
            return

        headers = {"Content-Type": "application/json"}

        # Check if the settings are already updated
        res = self.requests.get(self.index_url + "/_settings")
        try:
            res.raise_for_status()
        except requests.exceptions.HTTPError:
            logger.error("Error getting the index settings: {}".format(res.text))

        settings = res.json()
        index_settings = settings[self.index]['settings']['index']
        analysis = json.loads(analyzers)['settings']['analysis']
        if 'analysis' in index_settings and analysis == index_settings['analysis']:
            logger.debug("Index settings for {} is already updated. No need to update it".format(self.index))
            return

        close_index_url = "{}/_close".format(self.index_url)
        res = self.requests.post(close_index_url, headers=headers)
        try:
            res.raise_for_status()
        except requests.exceptions.HTTPError:
            logger.error("Error closing the index before updating settings: {}".format(res.text))

        url_set = "{}/_settings".format(self.index_url)
        res = self.requests.put(url_set, data=analyzers,
                                headers=headers)
        try:
            res.raise_for_status()
        except requests.exceptions.HTTPError:
            logger.error("Error updating index settings {}. Settings: {}".format(res.text, analyzers))

        open_index_url = "{}/_open".format(self.index_url)
        res = self.requests.post(open_index_url, headers=headers)
        try:
            res.raise_for_status()
            logger.debug("Index settings updated {}: {}".format(self.index, analyzers))
        except requests.exceptions.HTTPError:
            logger.error("Error opening the index after updating settings: {}".format(res.text))

    def create_mappings(self, mappings):
        """Create the mappings for a given index. It includes the index
        pattern plus dynamic templates.

        :param mappings: elastic_mapping.Mapping object
        """
        headers = {"Content-Type": "application/json"}

        for _type in mappings:

            url_map = self.get_mapping_url(_type)

            # First create the manual mappings
            if mappings[_type] != '{}':
                res = self.requests.put(url_map, data=mappings[_type],
                                        headers=headers)
                try:
                    res.raise_for_status()
                except requests.exceptions.HTTPError:
                    logger.error("Error creating ES mappings {}. Mapping: {}".format(res.text, str(mappings[_type])))

            # After version 6, strings are keywords (not analyzed)
            not_analyze_strings = """
            {
              "dynamic_templates": [
                { "notanalyzed": {
                      "match": "*",
                      "match_mapping_type": "string",
                      "mapping": {
                          "type": "keyword"
                      }
                   }
                },
                { "formatdate": {
                      "match": "*",
                      "match_mapping_type": "date",
                      "mapping": {
                          "type": "date",
                          "format" : "strict_date_optional_time||epoch_millis"
                      }
                   }
                }
              ]
            } """
            res = self.requests.put(url_map, data=not_analyze_strings, headers=headers)
            try:
                res.raise_for_status()
            except requests.exceptions.HTTPError:
                logger.error("Can't add mapping {}: {}".format(anonymize_url(url_map), not_analyze_strings))

    def get_last_date(self, field, filters_=[]):
        """Find the date of the last item stored in the index

        :param field: field with the data
        :param filters_: additional filters to find the date
        """
        last_date = self.get_last_item_field(field, filters_=filters_)

        return last_date

    def get_last_offset(self, field, filters_=[]):
        """Find the offset of the last item stored in the index

        :param field: field with the data
        :param filters_: additional filters to find the date
        """
        offset = self.get_last_item_field(field, filters_=filters_, offset=True)

        return offset

    def get_last_item_field(self, field, filters_=[], offset=False):
        """Find the offset/date of the last item stored in the index.

        :param field: field with the data
        :param filters_: additional filters to find the date
        :param offset: if True, returns the offset field instead of date field
        """
        last_value = None

        url = self.index_url
        url += "/_search"

        if filters_ is None:
            filters_ = []

        terms = []
        for filter_ in filters_:
            if not filter_:
                continue
            term = '''{"term" : { "%s" : "%s"}}''' % (filter_['name'], filter_['value'])
            terms.append(term)

        data_query = '''"query": {"bool": {"filter": [%s]}},''' % (','.join(terms))

        data_agg = '''
            "aggs": {
                "1": {
                  "max": {
                    "field": "%s"
                  }
                }
            }
        ''' % field

        data_json = '''
        { "size": 0, %s  %s
        } ''' % (data_query, data_agg)

        logger.debug("{} {}".format(anonymize_url(url), data_json))

        headers = {"Content-Type": "application/json"}

        res = self.requests.post(url, data=data_json, headers=headers)
        res.raise_for_status()
        res_json = res.json()

        if 'aggregations' in res_json:
            last_value = res_json["aggregations"]["1"]["value"]

            if offset:
                if last_value is not None:
                    last_value = int(last_value)
            else:
                if "value_as_string" in res_json["aggregations"]["1"]:
                    last_value = res_json["aggregations"]["1"]["value_as_string"]
                    last_value = str_to_datetime(last_value)
                else:
                    last_value = res_json["aggregations"]["1"]["value"]
                    if last_value:
                        try:
                            last_value = unixtime_to_datetime(last_value)
                        except InvalidDateError:
                            # last_value is in microsecs
                            last_value = unixtime_to_datetime(last_value / 1000)
        return last_value

    def delete_items(self, retention_time, time_field="metadata__updated_on"):
        """Delete documents updated before a given date

        :param retention_time: maximum number of minutes wrt the current date to retain the data
        :param time_field: time field to delete the data
        """
        if retention_time is None:
            logger.debug("[items retention] Retention policy disabled, no items will be deleted.")
            return

        if retention_time <= 0:
            logger.debug("[items retention] Minutes to retain must be greater than 0.")
            return

        before_date = get_diff_current_date(minutes=retention_time)
        before_date_str = before_date.isoformat()

        es_query = '''
                    {
                      "query": {
                        "range": {
                            "%s": {
                                "lte": "%s"
                            }
                        }
                      }
                    }
                    ''' % (time_field, before_date_str)

        r = self.requests.post(self.index_url + "/_delete_by_query?refresh",
                               data=es_query, headers=HEADER_JSON, verify=False)
        try:
            r.raise_for_status()
            r_json = r.json()
            logger.debug("[items retention] {} items deleted from {} before {}.".format(
                         r_json['deleted'], anonymize_url(self.index_url), before_date))
        except requests.exceptions.HTTPError as ex:
            logger.error("[items retention] Error deleted items from {}. {}".format(
                         anonymize_url(self.index_url), ex))
            return

    def all_properties(self):
        """Get all properties of a given index"""

        url = self.get_mapping_url(_type='items')
        r = self.requests.get(url, headers=HEADER_JSON, verify=False)
        try:
            r.raise_for_status()
            r_json = r.json()

            # ES 7.x
            properties = r_json[self.index]['mappings'].get('properties', {})

            # ES 6.x
            if not properties:
                items_mapping = r_json[self.index]['mappings'].get('items', {})
                properties = items_mapping.get('properties', {}) if items_mapping else {}

        except requests.exceptions.HTTPError as ex:
            logger.error("Error all attributes for {}. {}".format(anonymize_url(self.index_url), ex))
            return

        return properties

    def is_legacy(self):
        """ Simply calls the static version with it's own values """
        return ElasticSearch.is_legacy_static(self.major, self.distribution)

    @staticmethod
    def is_legacy_static(major, distribution):
        """ Returns true if ES < 7 or OS < 1, false otherwise.
        Static version exists because not every place that uses this check has an ES object."""
        if major is None:
            return False
        int_maj = int(major)
        return ((int_maj < 7 and distribution == 'elasticsearch')
                or (int_maj < 1 and distribution == 'opensearch'))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Alberto Pérez García-Plaza <alpgarcia@bitergia.com>
#

import logging
from datetime import datetime

import pandas
from elasticsearch import helpers, NotFoundError
from elasticsearch_dsl import Search, Q

from cereslib.enrich.enrich import Onion
from grimoirelab_toolkit import datetime as gl_dt
from grimoire_elk.enriched.ceres_base import ESConnector, CeresBase
from grimoire_elk.elastic import ElasticSearch


logger = logging.getLogger(__name__)


class ESOnionConnector(ESConnector):
    """ElasticSearch connector to ease data management related to Onion metric.

    Extends ES connector to work directly with Pandas DataFrames instead of using
    hits as they come from ES. This intends to ease the process of Onion metric.

    Writing is also ready to work directly with Pandas DataFrames.

    :param self._es_conn: ElasticSearch connection for reading from/writing to.
    :param self._es_index: ElasticSearch index for reading from/writing to.
    :param self._timeframe_field: date field to sort onion results.
    :param self._sort_on: date field to sort source index results, important for incremental process.
    :param self._read_only: True to avoid unwanted writes.
    """

    AUTHOR_NAME = 'author_name'
    AUTHOR_ORG = 'author_org_name'
    AUTHOR_MULTI_ORG_NAMES = 'author_multi_org_names'
    AUTHOR_UUID = 'author_uuid'
    CONTRIBUTIONS = 'contributions'
    LATEST_TS = 'latest_ts'
    TIMEFRAME = 'timeframe'
    TIMESTAMP = 'metadata__timestamp'
    PROJECT = 'project'

    def __init__(self, es_conn, es_index, contribs_field,
                 timeframe_field='grimoire_creation_date',
                 sort_on_field='metadata__timestamp', read_only=True):

        super().__init__(es_conn=es_conn, es_index=es_index, sort_on_field=sort_on_field, read_only=read_only)

        self.contribs_field = contribs_field
        self._timeframe_field = timeframe_field
        data_source = es_index.split("_")[0]
        self.__log_prefix = "[" + data_source + "] study onion"

    def read_block(self, size=None, from_date=None):
        """Read author commits by Quarter, Org and Project.

        :param from_date: not used here. Incremental mode not supported yet.
        :param size: not used here.
        :return: DataFrame with commit count per author, split by quarter, org and project.
        """
        # label to check whether the input index contains multiple affiliations. The label is
        # initialized to True, and can change its value only if no organizations exist in the
        # attribute `author_multi_org_names`
        multi_org = True

        # Get quarters corresponding to All items (Incremental mode NOT SUPPORTED)
        quarters = self.__quarters()

        for quarter in quarters:

            logger.info("{} Quarter: {}".format(self.__log_prefix, quarter))

            date_range = {self._timeframe_field: {'gte': quarter.start_time, 'lte': quarter.end_time}}

            orgs = self.__list_uniques(date_range, self.AUTHOR_MULTI_ORG_NAMES)
            if not orgs:
                multi_org = False
                logger.warning("{} Attribute {} not found, using {}".format(
                    self.__log_prefix, self.AUTHOR_MULTI_ORG_NAMES, self.AUTHOR_ORG)
                )
                orgs = self.__list_uniques(date_range, self.AUTHOR_ORG)

            projects = self.__list_uniques(date_range, self.PROJECT)

            # Get global data
            s = self.__build_search(date_range)
            response = s.execute()

            for timing in response.aggregations[self.TIMEFRAME].buckets:
                yield self.__build_dataframe(timing).copy()

            # Get global data by Org
            for org_name in orgs:

                logger.debug("{} Quarter: {}  Org: {}".format(self.__log_prefix, quarter, org_name))

                s = self.__build_search(date_range, org_name=org_name, multi_org=multi_org)
                response = s.execute()

                for timing in response.aggregations[self.TIMEFRAME].buckets:
                    yield self.__build_dataframe(timing, org_name=org_name).copy()

            # Get project specific data
            for project in projects:

                logger.debug("{} Quarter: {}  Project: {}".format(
                    self.__log_prefix, quarter, project))

                # Global project
                s = self.__build_search(date_range, project_name=project)
                response = s.execute()

                for timing in response.aggregations[self.TIMEFRAME].buckets:
                    yield self.__build_dataframe(timing, project_name=project).copy()

                # Split by Org
                for org_name in orgs:

                    logger.debug("{} Quarter: {}  Project: {}  Org: {}".format(
                                 self.__log_prefix, quarter, project, org_name))

                    s = self.__build_search(date_range, project_name=project, org_name=org_name, multi_org=multi_org)
                    response = s.execute()

                    for timing in response.aggregations[self.TIMEFRAME].buckets:
                        yield self.__build_dataframe(timing, project_name=project, org_name=org_name).copy()

    def write(self, items):
        """Write items into ElasticSearch.

        :param items: Pandas DataFrame
        """
        if self._read_only:
            raise IOError("Cannot write, Connector created as Read Only")

        if len(items) == 0:
            logger.info("{} Nothing to write".format(self.__log_prefix))
            return

        # Uploading info to the new ES
        rows = items.to_dict("index")
        docs = []
        for row_index in rows.keys():
            row = rows[row_index]
            item_id = row[self.AUTHOR_ORG] + '_' + row[self.PROJECT] + '_' \
                + row[self.TIMEFRAME] + '_' + row[self.AUTHOR_UUID]
            item_id = item_id.replace(' ', '').lower()

            doc = {
                "_index": self._es_index,
                "_type": "item",
                "_id": item_id,
                "_source": row
            }

            if not ElasticSearch.is_legacy_static(self._es_major, self._es_distribution):
                doc.pop('_type')

            docs.append(doc)

        # TODO uncomment following lines for incremental version
        # # Delete old data if exists to ensure refreshing in case of deleted commits
        # timeframe = docs[0]['_source']['timeframe']
        # org = docs[0]['_source']['author_org_name']
        # project = docs[0]['_source']['project']
        # s = Search(using=self._es_conn, index=self._es_index)
        # s = s.filter('term', project=project)
        # s = s.filter('term', author_org_name=org)
        # s = s.filter('term', timeframe=timeframe)
        # response = s.execute()
        #
        # if response.hits.total > 0:
        #     response = s.delete()
        #     logger.info("[Onion] Deleted " + str(response.deleted) + " items for refreshing: " + timeframe + " "
        #                 + org + " " + project)

        # TODO exception and error handling
        helpers.bulk(self._es_conn, docs)
        logger.debug("{} Written: {}".format(self.__log_prefix, len(docs)))

    def latest_enrichment_date(self):
        """Get the most recent enrichment date.

        :return: latest date based on `metadata__enriched_on` field,
                 None if no values found for that field.

        :raises NotFoundError: index not found in ElasticSearch
        """
        latest_date = None

        search = Search(using=self._es_conn, index=self._es_index)
        # from:to parameters (=> from: 0, size: 0)
        search = search[0:0]
        search = search.aggs.metric('max_date', 'max', field='metadata__enriched_on')

        try:
            response = search.execute()

            aggs = response.to_dict()['aggregations']
            if aggs['max_date']['value'] is None:
                logger.debug("{} No data for metadata__enriched_on field found in {} index".format(
                             self.__log_prefix, self._es_index))

            else:
                # Incremental case: retrieve items from last item in ES write index
                max_date = aggs['max_date']['value_as_string']
                latest_date = gl_dt.str_to_datetime(max_date)

        except NotFoundError as nfe:
            raise nfe

        return latest_date

    def __quarters(self, from_date=None):
        """Get a set of quarters with available items from a given index date.

        :param from_date:
        :return: list of `pandas.Period` corresponding to quarters
        """
        s = Search(using=self._es_conn, index=self._es_index)
        if from_date:
            # Work around to solve conversion problem of '__' to '.' in field name
            q = Q('range')
            q.__setattr__(self._sort_on_field, {'gte': from_date})
            s = s.filter(q)

        # from:to parameters (=> from: 0, size: 0)
        s = s[0:0]

        if ElasticSearch.is_legacy_static(self._es_major, self._es_distribution):
            s.aggs.bucket(self.TIMEFRAME, 'date_histogram', field=self._timeframe_field,
                          interval='quarter', min_doc_count=1)
        else:
            s.aggs.bucket(self.TIMEFRAME, 'date_histogram', field=self._timeframe_field,
                          calendar_interval='quarter', min_doc_count=1)
        response = s.execute()

        quarters = []
        for quarter in response.aggregations[self.TIMEFRAME].buckets:
            period = pandas.Period(quarter.key_as_string, 'Q')
            quarters.append(period)

        return quarters

    def __list_uniques(self, date_range, field_name):
        """Retrieve a list of unique values in a given field within a date range.

        :param date_range:
        :param field_name:
        :return: list  of unique values.
        """
        # Get project list
        s = Search(using=self._es_conn, index=self._es_index)
        s = s.filter('range', **date_range)
        # from:to parameters (=> from: 0, size: 0)
        s = s[0:0]
        s.aggs.bucket('uniques', 'terms', field=field_name, size=1000)
        response = s.execute()
        uniques_list = []
        for item in response.aggregations.uniques.buckets:
            uniques_list.append(item.key)

        return uniques_list

    def __build_search(self, date_range, project_name=None, org_name=None, multi_org=False):
        s = Search(using=self._es_conn, index=self._es_index)
        s = s.filter('range', **date_range)
        if project_name:
            s = s.filter('term', project=project_name)

        if org_name:
            if multi_org:
                s = s.filter('term', author_multi_org_names=org_name)
            else:
                s = s.filter('term', author_org_name=org_name)

        s = s.filter('term', author_bot='false')

        # from:to parameters (=> from: 0, size: 0)
        s = s[0:0]

        # Get author_name and most recent metadata__timestamp for quarter (should be enough per quarter,
        # computing it by user probably is not needed as we are going to recalculate the whole quarter)

        # We are not keeping all metadata__* fields because we are grouping commits by author, so we can only
        # store one value per author.
        if ElasticSearch.is_legacy_static(self._es_major, self._es_distribution):
            bucket = s.aggs.bucket(self.TIMEFRAME, 'date_histogram', field=self._timeframe_field, interval='quarter')
        else:
            bucket = s.aggs.bucket(self.TIMEFRAME, 'date_histogram', field=self._timeframe_field, calendar_interval='quarter')
        bucket.metric(self.LATEST_TS, 'max', field=self._sort_on_field) \
            .bucket(self.AUTHOR_UUID, 'terms', field=self.AUTHOR_UUID, size=1000) \
            .metric(self.CONTRIBUTIONS, 'cardinality', field=self.contribs_field, precision_threshold=40000)\
            .bucket(self.AUTHOR_NAME, 'terms', field=self.AUTHOR_NAME, size=1)
        return s

    def __build_dataframe(self, timing, project_name=None, org_name=None):
        """Build a DataFrame from a time bucket.

        :param timing:
        :param project_name:
        :param org_name:
        :return:
        """
        date_list = []
        uuid_list = []
        name_list = []
        contribs_list = []
        latest_ts_list = []
        logger.debug("{} timing: {}".format(self.__log_prefix, timing.key_as_string))

        for author in timing[self.AUTHOR_UUID].buckets:
            latest_ts_list.append(timing[self.LATEST_TS].value_as_string)
            date_list.append(timing.key_as_string)
            uuid_list.append(author.key)
            if author[self.AUTHOR_NAME] and author[self.AUTHOR_NAME].buckets \
                    and len(author[self.AUTHOR_NAME].buckets) > 0:
                name_list.append(author[self.AUTHOR_NAME].buckets[0].key)
            else:
                name_list.append("Unknown")
            contribs_list.append(author[self.CONTRIBUTIONS].value)

        df = pandas.DataFrame()
        df[self.TIMEFRAME] = date_list
        df[self.AUTHOR_UUID] = uuid_list
        df[self.AUTHOR_NAME] = name_list
        df[self.CONTRIBUTIONS] = contribs_list
        df[self.TIMESTAMP] = latest_ts_list

        if not project_name:
            project_name = "_Global_"
        df[self.PROJECT] = project_name

        if not org_name:
            org_name = "_Global_"
        df[self.AUTHOR_ORG] = org_name

        return df


class OnionStudy(CeresBase):
    """Compute Onion metric on a Git enriched index.

    Generate one entry per each modified file involved in each commit, storing some specific fields
    like file name, path, path parts and extension.

    :param self._in: ESOnionConnector for reading source items from.
    :param self._out: ESOnionConnector to write processed items to.
    """

    def __init__(self, in_connector, out_connector, data_source):

        super().__init__(in_connector, out_connector, None)

        self.data_source = data_source
        self.__log_prefix = "[" + data_source + "] study onion"

    def process(self, items_block):
        """Process a DataFrame to compute Onion.

        :param items_block: items to be processed. Expects to find a pandas DataFrame.
        """

        logger.debug("{} Authors to process: {}".format(self.__log_prefix, len(items_block)))

        onion_enrich = Onion(items_block)
        df_onion = onion_enrich.enrich(member_column=ESOnionConnector.AUTHOR_UUID,
                                       events_column=ESOnionConnector.CONTRIBUTIONS)

        # Get and store Quarter as String
        df_onion['quarter'] = df_onion[ESOnionConnector.TIMEFRAME].map(lambda x: str(pandas.Period(x, 'Q')))

        # Add metadata: enriched on timestamp
        df_onion['metadata__enriched_on'] = datetime.utcnow().isoformat()
        df_onion['data_source'] = self.data_source
        df_onion['grimoire_creation_date'] = df_onion[ESOnionConnector.TIMEFRAME]

        logger.debug("{} Final new events: {}".format(self.__log_prefix, len(df_onion)))

        return self.ProcessResults(processed=len(df_onion), out_items=df_onion)


def onion_study(in_conn, out_conn, data_source):
    """Build and index for onion from a given Git index.

    :param in_conn: ESPandasConnector to read from.
    :param out_conn: ESPandasConnector to write to.
    :param data_source: name of the date source to generate onion from.
    :return: number of documents written in ElasticSearch enriched index.
    """
    onion = OnionStudy(in_connector=in_conn, out_connector=out_conn, data_source=data_source)
    ndocs = onion.analyze()
    return ndocs
